""" module irafinst.py - Defines bare-bones functionality needed when IRAF
is not present.  This is NOT a replacement for any major parts of IRAF.
For now, in general, we assume that IRAF exists until we are told otherwise.

Obviously, this module should refrain as much as possible from importing any
IRAF related code (at least globally), since this is heavily relied upon in
non-IRAF situations.
"""
import os
import sys

# File name prefix signal
NO_IRAF_PFX = f'{os.path.dirname(__file__)}/noiraf'

# Are we running without an IRAF installation?  If no, EXISTS == False
if sys.platform.startswith('win'):
    EXISTS = False
else:
    EXISTS = 'PYRAF_NO_IRAF' not in os.environ


def getIrafVer():
    """ Return current IRAF version as a string """
    from . import iraffunctions
    cltask = iraffunctions.getTask('cl')
    # must use default par list, in case they have a local override
    plist = cltask.getDefaultParList()
    # get the 'release' par and then get it's value
    release = [p.value for p in plist if p.name == 'release']
    return release[0]  # only 1 item in list


def getIrafVerTup():
    """ Return current IRAF version as a tuple (ints until last item) """
    verlist = getIrafVer().split('.')
    outlist = []
    for v in verlist:
        if v.isdigit():
            outlist.append(int(v))
        else:
            outlist.append(v)
    return tuple(outlist)



import pytest

from ..minmatch import MinMatchDict, AmbiguousKeyError

BASEKEYS = tuple(['test', 'text', 'ten'])
BASEVALUES = tuple([1, 2, 10])


@pytest.fixture
def mmd():
    d = MinMatchDict()
    for value in zip(*[BASEKEYS, BASEVALUES]):
        d.add(*value)
    return d


def test_ambiguous_assignment_key(mmd):
    with pytest.raises(AmbiguousKeyError):
        mmd['te'] = 5


def test_ambiguous_assignment_get_t(mmd):
    with pytest.raises(AmbiguousKeyError):
        mmd.get('t')


def test_ambiguous_assignment_get_tes(mmd):
    with pytest.raises(AmbiguousKeyError):
        mmd.get('te')


def test_ambiguous_assignment_del_tes(mmd):
    with pytest.raises(AmbiguousKeyError):
        del mmd['te']


def test_invalid_key_assignment(mmd):
    with pytest.raises(KeyError):
        mmd['t']


def test_dict_sort(mmd):
    result = [key for key, _ in sorted(mmd.items())]
    assert result[0] == 'ten'
    assert result[-1] == 'text'


@pytest.mark.parametrize('key', BASEKEYS)
def test_get_values(mmd, key):
    assert mmd.get(key)


def test_missing_key_returns_none(mmd):
    assert mmd.get('teq') is None


def test_getall(mmd):
    return mmd.getall('t')


def test_getall_returns_expected_values(mmd):
    result = mmd.getall('t')
    assert sorted(result) == [x for x in BASEVALUES]


def test_del_key(mmd):
    del mmd['test']


def test_del_keys(mmd):
    for key in BASEKEYS:
        del mmd[key]


def test_clear_dict(mmd):
    mmd.clear()
    assert mmd == dict()


def test_has_key(mmd):
    for key in BASEKEYS:
        # Ditch last character in string
        if len(key) > 3:
            key = key[:-1]

        assert mmd.has_key(key, exact=False)


def test_has_key_exact(mmd):
    for key in BASEKEYS:
        assert mmd.has_key(key, exact=True)


def test_key_in_dict(mmd):
    for key in BASEKEYS:
        assert key in mmd


def test_update_dict(mmd):
    new_dict = dict(ab=0)
    mmd.update(new_dict)
    assert 'test' in mmd and 'ab' in mmd



""" Contains the TaskPars class and any related functions.

$Id$
"""


class NoExecError(Exception):
    pass


class TaskPars:
    """ This represents a task's collection of configurable parameters.
    This class is meant to be mostly abstract, though there is some
    functionality included which could be common to most derived classes.
    This also serves to document the interface which must be met for EPAR.
    """

    def getName(self, *args, **kw):
        """ Returns the string name of the task. """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def getPkgname(self, *args, **kw):
        """ Returns the string name of the package, if applicable. """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def getParList(self, *args, **kw):
        """ Returns a list of parameter objects. """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def getDefaultParList(self, *args, **kw):
        """ Returns a list of parameter objects with default values set. """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def setParam(self, *args, **kw):
        """ Allows one to set the value of a single parameter.
            Initial signature is setParam(name, value, scope='', check=1) """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def getFilename(self, *args, **kw):
        """ Returns the string name of any associated config/parameter file. """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def saveParList(self, *args, **kw):
        """ Allows one to save the entire set to a file. """
        raise NotImplementedError("class TaskPars is not to be used directly")

    def run(self, *args, **kw):
        """ Runs the task with the known parameters. """
        raise NoExecError("Bug: class TaskPars is not to be used directly")

    def canPerformValidation(self):
        """ Returns bool.  If True, expect tryValue() to be called next. """
        return False

    def knowAsNative(self):
        """ Returns bool.  Return true if the class prefers in-memory objects
        to keep (know) their parameter values in native format instead of as
        strings. """
        return False

    def getHelpAsString(self):
        """ Meant to be overridden - return a task specific help string. """
        return 'No help string available for task "'+self.getName()+'".\n '+ \
               'Implement getHelpAsString() in your TaskPars sub-class.'

    # also, eparam, lParam, tParam, dParam, tryValue ?



"""
Small scanf implementation.

Python has powerful regular expressions but sometimes they are totally overkill
when you just want to parse a simple-formatted string.
C programmers use the scanf-function for these tasks (see link below).

This implementation of scanf translates the simple scanf-format into
regular expressions. Unlike C you can be sure that there are no buffer overflows
possible.

For more information see
  * http://www.python.org/doc/current/lib/node49.html
  * http://en.wikipedia.org/wiki/Scanf

Original code from:
    https://github.com/joshburnett/scanf (version 1.5.2)

Modified for the needs of PyRAF:
 * all fields may have a max width (not a fixed width)
 * add "l" (for [outdated] long ints)
 * allow "0x" and "0" prefixes in ints for hexa/octal numbers

Differences to the original PyRAF sscanf module:
 * "n" coversion missing (number of characters so far)


"""
import re
import sys
try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

__version__ = '1.5.2'

__all__ = ["scanf", 'scanf_translate', 'scanf_compile']


DEBUG = False

def sint(s):
    if s.startswith("0x"):
        return int(s[2:], 16)
    elif s.startswith("0"):
        return int(s[1:], 8)
    else:
        return int(s)

# As you can probably see it is relatively easy to add more format types.
# Make sure you add a second entry for each new item that adds the extra
#   few characters needed to handle the field ommision.
scanf_translate = [
    (re.compile(_token), _pattern, _cast) for _token, _pattern, _cast in [
        (r"%c", r"(.)", lambda x:x),
        (r"%\*c", r"(?:.)", None),

        (r"%(\d+)c", r"(.{0,%s})", lambda x:x),
        (r"%\*(\d+)c", r"(?:.{0,%s})", None),

        (r"%s", r"(\S+)", lambda x: x),
        (r"%\*s", r"(?:\S+)", None),

        (r"%(\d+)s", r"(\S{1,%s})", lambda x:x),
        (r"%\*(\d+)s", r"(?:\S{1,%s})", None),

        (r"%\[([^\]]+)\]", r"([%s]+)", lambda x:x),
        (r"%\*\[([^\]]+)\]", r"(?:[%s]+)", None),

        (r"%l?[dil]", r"([+-]?\d+)", sint),
        (r"%\*l?[dil]", r"(?:[+-]?\d+)", None),

        (r"%(\d+)l?[dil]", r"([+-]?\d{1,%s})", sint),
        (r"%\*(\d+)l?[dil]", r"(?:[+-]?\d{1,%s})", None),

        (r"%l?u", r"(\d+)", int),
        (r"%\*l?u", r"(?:\d+)", None),

        (r"%(\d+)l?u", r"(\d{1,%s})", int),
        (r"%\*(\d+)l?u", r"(?:\d{1,%s})", None),

        (r"%[fgeE]", r"([-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)", float),
        (r"%\*[fgeE]", r"(?:[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)", None),

        (r"%(\d+)[fgeE]", r"([-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)", float),
        (r"%\*(\d+)[fgeE]", r"(?:[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)", None),

        (r"%l?[xX]", r"((?:0[xX])?[\dA-Za-f]+)", lambda x: int(x, 16)),
        (r"%\*l?[xX]", r"(?:(?:0[xX])?[\dA-Za-f]+)", None),

        (r"%(\d+)l?[xX]", r"((?:0[xX])?[\dA-Za-f]{1,%s})", lambda x: int(x, 16)),
        (r"%\*(\d+)l?[xX]", r"(?:(?:0[xX])?[\dA-Za-f]{1,%s})", None),

        (r"%l?o", r"([0-7]*)", lambda x:int(x, 8)),
        (r"%\*l?o", r"(?:[0-7]*)", None),

        (r"%(\d+)l?o", r"([0-7]{1,%s})", lambda x: int(x, 8)),
        (r"%\*(\d+)l?o", r"(?:[0-7]{1,%s})", None),
    ]]


# Cache formats
SCANF_CACHE_SIZE = 1000


@lru_cache(maxsize=SCANF_CACHE_SIZE)
def scanf_compile(format, collapseWhitespace=True):
    """
    Translate the format into a regular expression

    For example:

        >>> format_re, casts = scanf_compile('%s - %d errors, %d warnings')
        >>> print format_re.pattern
        (\\S+) \- ([+-]?\\d+) errors, ([+-]?\\d+) warnings

    Translated formats are cached for faster reuse
    """

    format_pat = ""
    cast_list = []
    i = 0
    length = len(format)
    while i < length:
        found = None
        for token, pattern, cast in scanf_translate:
            found = token.match(format, i)
            if found:
                if cast: # cast != None
                    cast_list.append(cast)
                groups = found.groupdict() or found.groups()
                if groups:
                    pattern = pattern % groups
                format_pat += pattern
                i = found.end()
                break
        if not found:
            char = format[i]
            # escape special characters
            if char in "|^$()[]-.+*?{}<>\\":
                format_pat += "\\"
            format_pat += char
            i += 1
    if DEBUG:
        print("DEBUG: %r -> %s" % (format, format_pat))
    if collapseWhitespace:
        format_pat = re.sub(r'\s+', r'\\s+', format_pat)

    format_re = re.compile(format_pat)
    return format_re, cast_list


def scanf(format, s=None, collapseWhitespace=True):
    """Conversion specification are of the form:

        %[*][<max_width>]['l']<type_character>.

    The following format conversions are supported:

    %c            Fixed width character string.
    %s            String of non-whitespace characters with leading
                     whitespace skipped.
    %d, %i, %l    Signed integer (leading 0 => octal, 0x => hex).
    %o            Octal integer.
    %u            Unsigned integer.
    %x            Hexadecimal integer.
    %f, %g, %e    Python float
    %[]           Character scan set

    scanf.scanf returns a tuple of found values or None if the format
    does not match.

    """

    if s is None:
        s = sys.stdin

    if hasattr(s, "readline"):
        s = s.readline()

    format_re, casts = scanf_compile(format, collapseWhitespace)

    found = format_re.search(s)
    if found:
        groups = found.groups()
        return tuple([casts[i](groups[i]) for i in range(len(groups))])



"""basicpar.py -- General base class for parameter objects.  Broken out
                  from PyRAF's IrafPar class.

$Id$
"""
import re
import sys
from . import irafutils, minmatch
from .irafglobals import INDEF, Verbose, yes, no, clFloat

int_types = (int, )


# container class used for __deepcopy__ method
class _EmptyClass:
    pass


# -----------------------------------------------------
# Warning (non-fatal) error.  Raise an exception if in
# strict mode, or print a message if Verbose is on.
# -----------------------------------------------------

# Verbose (set irafglobals.py) determines
# whether warning messages are printed when errors are found.  The
# strict parameter to various methods and functions can be set to
# raise an exception on errors; otherwise we do our best to work
# around errors, only raising an exception for really serious,
# unrecoverable problems.

def warning(msg, strict=0, exception=SyntaxError, level=0):
    if strict:
        raise exception(msg)
    elif Verbose>level:
        sys.stdout.flush()
        sys.stderr.write('Warning: %s' % msg)
        if msg[-1:] != '\n': sys.stderr.write('\n')

# -----------------------------------------------------
# basic parameter factory
# -----------------------------------------------------

_string_types = [ 's', 'f', 'struct', 'z' ]
_real_types = [ 'r', 'd' ]

def parFactory(fields, strict=0):

    """parameter factory function

    fields is a list of the comma-separated fields (as in the .par file).
    Each entry is a string or None (indicating that field was omitted.)

    Set the strict parameter to a non-zero value to do stricter parsing
    (to find errors in the input)"""

    if len(fields) < 3 or None in fields[0:3]:
        raise SyntaxError("At least 3 fields must be given")
    type = fields[1]
    if type in _string_types:
        return IrafParS(fields,strict)
    elif type == 'R':
        return StrictParR(fields,1)
    elif type in _real_types:
        return IrafParR(fields,strict)
    elif type == "I":
        return StrictParI(fields,1)
    elif type == "i":
        return IrafParI(fields,strict)
    elif type == "b":
        return IrafParB(fields,strict)
    elif type == "ar":
        return IrafParAR(fields,strict)
    elif type == "ai":
        return IrafParAI(fields,strict)
    elif type == "as":
        return IrafParAS(fields,strict)
    elif type == "ab":
        return IrafParAB(fields,strict)
    elif type[:1] == "a":
        raise SyntaxError("Cannot handle arrays of type %s" % type)
    else:
        raise SyntaxError("Cannot handle parameter type %s" % type)


# --------------------------------------------------------
# Publish the (simple) algorithm for combining scope+name
# --------------------------------------------------------

def makeFullName(parScope, parName):
    """ Create the fully-qualified name (inclues scope if used) """
    # Skip scope (and leading dot) if no scope, even in cases where scope
    # IS used for other pars in the same task.
    if parScope:
        return parScope+'.'+parName
    else:
        return parName

# -----------------------------------------------------
# Set up minmatch dictionaries for parameter fields
# -----------------------------------------------------

flist = ("p_name", "p_xtype", "p_type", "p_mode", "p_prompt", "p_scope",
         "p_value", "p_default", "p_filename", "p_maximum", "p_minimum")
_getFieldDict = minmatch.MinMatchDict()
for field in flist: _getFieldDict.add(field, field)

flist = ("p_prompt", "p_value", "p_filename", "p_maximum", "p_minimum", "p_mode", "p_scope")
_setFieldDict = minmatch.MinMatchDict()
for field in flist: _setFieldDict.add(field, field)
del flist, field

# utility function to check whether string is a parameter field

def isParField(s):
    """Returns true if string s appears to be a parameter field"""
    try:
        return (s[:2] == "p_") and s in _getFieldDict
    except minmatch.AmbiguousKeyError:
        # If ambiguous match, assume it is a parameter field.
        # An exception will doubtless be raised later, but
        # there's really no good choice here.
        return 1

# basic IrafPar attributes
# IrafPar's are protected in setattr against adding arbitrary attributes,
# and this dictionary is used as a helper in instance initialization
_IrafPar_attr_dict = {
        "name" : None,
        "type" : None,
        "mode" : None,
        "value" : None,
        "min" : None,
        "max" : None,
        "choice" : None,
        "choiceDict" : None,
        "prompt" : None,
        "flags" : 0,
        "scope" : None,
        }

# flag bits tell whether value has been changed and
# whether it was set on the command line.
_changedFlag = 1
_cmdlineFlag = 2

# -----------------------------------------------------
# IRAF parameter base class
# -----------------------------------------------------

class IrafPar:

    """Non-array IRAF parameter base class"""

    def __init__(self,fields,strict=0):
        orig_len = len(fields)
        if orig_len < 3 or None in fields[0:3]:
            raise SyntaxError("At least 3 fields must be given")
        #
        # all the attributes that are going to get defined
        #
        self.__dict__.update(_IrafPar_attr_dict)
        self.name   = fields[0]
        self.type   = fields[1]
        self.mode   = fields[2]
        self.scope  = None # simple default; may be unused
        #
        # put fields into appropriate attributes
        #
        while len(fields) < 7: fields.append(None)
        #
        self.value = self._coerceValue(fields[3],strict)
        if fields[4] is not None and '|' in fields[4]:
            self._setChoice(fields[4].strip(),strict)
            if fields[5] is not None:
                if orig_len < 7:
                    warning("Max value illegal when choice list given" +
                                    " for parameter " + self.name +
                                    " (probably missing comma)",
                                    strict)
                    # try to recover by assuming max string is prompt
                    fields[6] = fields[5]
                    fields[5] = None
                else:
                    warning("Max value illegal when choice list given" +
                            " for parameter " + self.name, strict)
        else:
            #XXX should catch ValueError exceptions here and set to null
            #XXX could also check for missing comma (null prompt, prompt
            #XXX in max field)
            if fields[4] is not None:
                self.min = self._coerceValue(fields[4],strict)
            if fields[5] is not None:
                self.max = self._coerceValue(fields[5],strict)
        if self.min not in [None, INDEF] and \
           self.max not in [None, INDEF] and self.max < self.min:
            warning("Max " + str(self.max) + " is less than minimum " + \
                    str(self.min) + " for parameter " + self.name,
                    strict)
            self.min, self.max = self.max, self.min
        if fields[6] is not None:
            self.prompt = irafutils.removeEscapes(
                                            irafutils.stripQuotes(fields[6]))
        else:
            self.prompt = ''
        #
        # check attributes to make sure they are appropriate for
        # this parameter type (e.g. some do not allow choice list
        # or min/max)
        #
        self._checkAttribs(strict)
        #
        # check parameter value to see if it is correct
        #
        try:
            self.checkValue(self.value,strict)
        except ValueError as e:
            warning("Illegal initial value for parameter\n" + str(e),
                    strict, exception=ValueError)
            # Set illegal values to None, just like IRAF
            self.value = None

    #--------------------------------------------
    # public accessor methods
    #--------------------------------------------

    def isLegal(self):
        """Returns true if current parameter value is legal"""
        try:
            # apply a stricter definition of legal here
            # fixable values have already been fixed
            # don't accept None values
            self.checkValue(self.value)
            return self.value is not None
        except ValueError:
            return 0

    def setScope(self,value=''):
        """Set scope value.  Written this way to not change the
           standard set of fields in the comma-separated list. """
        # set through dictionary to avoid extra calls to __setattr__
        self.__dict__['scope'] = value

    def setCmdline(self,value=1):
        """Set cmdline flag"""
        # set through dictionary to avoid extra calls to __setattr__
        if value:
            self.__dict__['flags'] = self.flags | _cmdlineFlag
        else:
            self.__dict__['flags'] = self.flags & ~_cmdlineFlag

    def isCmdline(self):
        """Return cmdline flag"""
        return (self.flags & _cmdlineFlag) == _cmdlineFlag

    def setChanged(self,value=1):
        """Set changed flag"""
        # set through dictionary to avoid another call to __setattr__
        if value:
            self.__dict__['flags'] = self.flags | _changedFlag
        else:
            self.__dict__['flags'] = self.flags & ~_changedFlag

    def isChanged(self):
        """Return changed flag"""
        return (self.flags & _changedFlag) == _changedFlag

    def setFlags(self,value):
        """Set all flags"""
        self.__dict__['flags'] = value

    def isLearned(self, mode=None):
        """Return true if this parameter is learned

        Hidden parameters are not learned; automatic parameters inherit
        behavior from package/cl; other parameters are learned.
        If mode is set, it determines how automatic parameters behave.
        If not set, cl.mode parameter determines behavior.
        """
        if "l" in self.mode: return 1
        if "h" in self.mode: return 0
        if "a" in self.mode:
            if mode is None: mode = 'ql' # that is, iraf.cl.mode
            if "h" in mode and "l" not in mode:
                return 0
        return 1

    #--------------------------------------------
    # other public methods
    #--------------------------------------------

    def getPrompt(self):
        """Alias for getWithPrompt() for backward compatibility"""
        return self.getWithPrompt()

    def getWithPrompt(self):
        """Interactively prompt for parameter value"""
        if self.prompt:
            pstring = self.prompt.split("\n")[0].strip()
        else:
            pstring = self.name
        if self.choice:
            schoice = list(map(self.toString, self.choice))
            pstring = pstring + " (" + "|".join(schoice) + ")"
        elif self.min not in [None, INDEF] or \
                 self.max not in [None, INDEF]:
            pstring = pstring + " ("
            if self.min not in [None, INDEF]:
                pstring = pstring + self.toString(self.min)
            pstring = pstring + ":"
            if self.max not in [None, INDEF]:
                pstring = pstring + self.toString(self.max)
            pstring = pstring + ")"
        # add current value as default
        if self.value is not None:
            pstring = pstring + " (" + self.toString(self.value,quoted=1) + ")"
        pstring = pstring + ": "
        # don't redirect stdin/out unless redirected filehandles are also ttys
        # or unless originals are NOT ttys
        stdout = sys.__stdout__
        try:
            if sys.stdout.isatty() or not stdout.isatty():
                stdout = sys.stdout
        except AttributeError:
            pass
        stdin = sys.__stdin__
        try:
            if sys.stdin.isatty() or not stdin.isatty():
                stdin = sys.stdin
        except AttributeError:
            pass
        # print prompt, suppressing both newline and following space
        stdout.write(pstring)
        stdout.flush()
        ovalue = irafutils.tkreadline(stdin)
        value = ovalue.strip()
        # loop until we get an acceptable value
        while (1):
            try:
                # null input usually means use current value as default
                # check it anyway since it might not be acceptable
                if value == "": value = self._nullPrompt()
                self.set(value)
                # None (no value) is not acceptable value after prompt
                if self.value is not None: return
                # if not EOF, keep looping
                if ovalue == "":
                    stdout.flush()
                    raise EOFError("EOF on parameter prompt")
                print("Error: specify a value for the parameter")
            except ValueError as e:
                print(str(e))
            stdout.write(pstring)
            stdout.flush()
            ovalue = irafutils.tkreadline(stdin)
            value = ovalue.strip()

    def get(self, field=None, index=None, lpar=0, prompt=1, native=0, mode="h"):
        """Return value of this parameter as a string (or in native format
        if native is non-zero.)"""

        if field and field != "p_value":
            # note p_value comes back to this routine, so shortcut that case
            return self._getField(field,native=native,prompt=prompt)

        # may prompt for value if prompt flag is set
        if prompt: self._optionalPrompt(mode)

        if index is not None:
            raise SyntaxError("Parameter "+self.name+" is not an array")

        if native:
            rv = self.value
        else:
            rv = self.toString(self.value)
        return rv

    def set(self, value, field=None, index=None, check=1):
        """Set value of this parameter from a string or other value.
        Field is optional parameter field (p_prompt, p_minimum, etc.)
        Index is optional array index (zero-based).  Set check=0 to
        assign the value without checking to see if it is within
        the min-max range or in the choice list."""

        if index is not None:
            raise SyntaxError("Parameter "+self.name+" is not an array")

        if field:
            self._setField(value,field,check=check)
        else:
            if check:
                self.value = self.checkValue(value)
            else:
                self.value = self._coerceValue(value)
            self.setChanged()

    def checkValue(self,value,strict=0):
        """Check and convert a parameter value.

        Raises an exception if the value is not permitted for this
        parameter.  Otherwise returns the value (converted to the
        right type.)
        """
        v = self._coerceValue(value,strict)
        return self.checkOneValue(v,strict)

    def checkOneValue(self,v,strict=0):
        """Checks a single value to see if it is in range or choice list

        Allows indirection strings starting with ")".  Assumes
        v has already been converted to right value by
        _coerceOneValue.  Returns value if OK, or raises
        ValueError if not OK.
        """
        if v in [None, INDEF] or (isinstance(v,str) and v[:1] == ")"):
            return v
        elif v == "":
            # most parameters treat null string as omitted value
            return None
        elif self.choice is not None and v not in self.choiceDict:
            schoice = list(map(self.toString, self.choice))
            schoice = "|".join(schoice)
            raise ValueError("Parameter %s: "
                    "value %s is not in choice list (%s)" %
                    (self.name, str(v), schoice))
        elif (self.min not in [None, INDEF] and v<self.min):
            raise ValueError("Parameter %s: "
                    "value `%s' is less than minimum `%s'" %
                    (self.name, str(v), str(self.min)))
        elif (self.max not in [None, INDEF] and v>self.max):
            raise ValueError("Parameter %s: "
                    "value `%s' is greater than maximum `%s'" %
                    (self.name, str(v), str(self.max)))
        return v

    def dpar(self, cl=1):
        """Return dpar-style executable assignment for parameter

        Default is to write CL version of code; if cl parameter is
        false, writes Python executable code instead.
        """
        sval = self.toString(self.value, quoted=1)
        if not cl:
            if sval == "": sval = "None"
        s = "%s = %s" % (self.name, sval)
        return s

    def fullName(self):
        """ Return the fully-qualified name (inclues scope if used) """
        return makeFullName(self.scope, self.name) # scope can be None or ''

    def pretty(self,verbose=0):
        """Return pretty list description of parameter"""
        # split prompt lines and add blanks in later lines to align them
        plines = self.prompt.split('\n')
        for i in range(len(plines)-1): plines[i+1] = 32*' ' + plines[i+1]
        plines = '\n'.join(plines)
        namelen = min(len(self.name), 12)
        pvalue = self.get(prompt=0,lpar=1)
        alwaysquoted = ['s', 'f', '*gcur', '*imcur', '*ukey', 'pset']
        if self.type in alwaysquoted and self.value is not None: pvalue = '"' + pvalue + '"'
        if self.mode == "h":
            s = "%13s = %-15s %s" % ("("+self.name[:namelen],
                                    pvalue+")", plines)
        else:
            s = "%13s = %-15s %s" % (self.name[:namelen],
                                    pvalue, plines)
        if not verbose: return s

        if self.choice is not None:
            s = s + "\n" + 32*" " + "|"
            nline = 33
            for i in range(len(self.choice)):
                sch = str(self.choice[i]) + "|"
                s = s + sch
                nline = nline + len(sch) + 1
                if nline > 80:
                    s = s + "\n" + 32*" " + "|"
                    nline = 33
        elif self.min not in [None, INDEF] or self.max not in [None, INDEF]:
            s = s + "\n" + 32*" "
            if self.min not in [None, INDEF]:
                s = s + str(self.min) + " <= "
            s = s + self.name
            if self.max not in [None, INDEF]:
                s = s + " <= " + str(self.max)
        return s

    def save(self, dolist=0):
        """Return .par format string for this parameter

        If dolist is set, returns fields as a list of strings.  Default
        is to return a single string appropriate for writing to a file.
        """
        quoted = not dolist
        fields = 7*[""]
        fields[0] = self.name
        fields[1] = self.type
        fields[2] = self.mode
        fields[3] = self.toString(self.value,quoted=quoted)
        if self.choice is not None:
            schoice = list(map(self.toString, self.choice))
            schoice.insert(0,'')
            schoice.append('')
            fields[4] = repr('|'.join(schoice))
        elif self.min not in [None,INDEF]:
            fields[4] = self.toString(self.min,quoted=quoted)
        if self.max not in [None,INDEF]:
            fields[5] = self.toString(self.max,quoted=quoted)
        if self.prompt:
            if quoted:
                sprompt = repr(self.prompt)
            else:
                sprompt = self.prompt
            # prompt can have embedded newlines (which are printed)
            sprompt = sprompt.replace(r'\012', '\n')
            sprompt = sprompt.replace(r'\n', '\n')
            fields[6] = sprompt
        # delete trailing null parameters
        for i in [6,5,4]:
            if fields[i] != "": break
            del fields[i]
        if dolist:
            return fields
        else:
            return ','.join(fields)

    #--------------------------------------------
    # special methods to give desired object syntax
    #--------------------------------------------

    # allow parameter object to be used in arithmetic expression

    def __coerce__(self, other):
        return coerce(self.get(native=1), other)

    # fields are accessible as attributes

    def __getattr__(self,field):
        if field[:1] == '_':
            raise AttributeError(field)
        try:
            return self._getField(field, native=1)
        except SyntaxError as e:
            if field in _IrafPar_attr_dict:
                # handle odd-ball case of new code accessing par's new
                # attr (e.g. scope), with old-code-cached version of par
                return _IrafPar_attr_dict[field] # return unused default
            else:
                raise AttributeError(str(e))

    def __setattr__(self,attr,value):
        # don't allow any new parameters to be added
        if attr in self.__dict__:
            self.__dict__[attr] = value
        elif isParField(attr):
            #XXX should check=0 be used here?
            self._setField(value, attr)
        else:
            raise AttributeError("No attribute %s for parameter %s" %
                    (attr, self.name))

    def __deepcopy__(self, memo):
        """Deep copy of this parameter object"""
        new = _EmptyClass()
        # shallow copy of dictionary suffices for most attributes
        new.__dict__ = self.__dict__.copy()
        # value, choice may be lists of atomic items
        if isinstance(self.value, list):
            new.value = list(self.value)
        if isinstance(self.choice, list):
            new.choice = list(self.choice)
        # choiceDict is OK with shallow copy because it will
        # always be reset if choices change
        new.__class__ = self.__class__
        return new

    def __getstate__(self):
        """Return state info for pickle"""
        # choiceDict gets reconstructed
        if self.choice is None:
            return self.__dict__
        else:
            d = self.__dict__.copy()
            d['choiceDict'] = None
            return d

    def __setstate__(self, state):
        """Restore state info from pickle"""
        self.__dict__.clear()
        self.__dict__.update(state)
        if self.choice is not None:
            self._setChoiceDict()

    def __str__(self):
        """Return readable description of parameter"""
        s = "<" + self.__class__.__name__ + " " + self.name + " " + self.type
        s = s + " " + self.mode + " " + repr(self.value)
        if self.choice is not None:
            schoice = list(map(self.toString, self.choice))
            s = s + " |" + "|".join(schoice) + "|"
        else:
            s = s + " " + repr(self.min) + " " + repr(self.max)
        s = s + ' "' + self.prompt + '">'
        return s

    #--------------------------------------------
    # private methods -- may be used by subclasses, but should
    # not be needed outside this module
    #--------------------------------------------

    def _checkAttribs(self,strict=0):
        # by default no restrictions on attributes
        pass

    def _setChoice(self,s,strict=0):
        """Set choice parameter from string s"""
        clist = _getChoice(s,strict)
        self.choice = list(map(self._coerceValue, clist))
        self._setChoiceDict()

    def _setChoiceDict(self):
        """Create dictionary for choice list"""
        # value is name of choice parameter (same as key)
        self.choiceDict = {}
        for c in self.choice: self.choiceDict[c] = c

    def _nullPrompt(self):
        """Returns value to use when answer to prompt is null string"""
        # most parameters just keep current default (even if None)
        return self.value

    def _optionalPrompt(self, mode):
        """Interactively prompt for parameter if necessary

        Prompt for value if
        (1) mode is hidden but value is undefined or bad, or
        (2) mode is query and value was not set on command line
        Never prompt for "u" mode parameters, which are local variables.
        """
        if (self.mode == "h") or (self.mode == "a" and mode == "h"):
            # hidden parameter
            if not self.isLegal():
                self.getWithPrompt()
        elif self.mode == "u":
            # "u" is a special mode used for local variables in CL scripts
            # They should never prompt under any circumstances
            if not self.isLegal():
                raise ValueError(
                                "Attempt to access undefined local variable `%s'" %
                                self.name)
        else:
            # query parameter
            if self.isCmdline()==0:
                self.getWithPrompt()

    def _getPFilename(self,native,prompt):
        """Get p_filename field for this parameter

        Same as get for non-list params
        """
        return self.get(native=native,prompt=prompt)

    def _getPType(self):
        """Get underlying datatype for this parameter

        Just self.type for normal params
        """
        return self.type

    def _getField(self, field, native=0, prompt=1):
        """Get a parameter field value"""
        try:
            # expand field name using minimum match
            field = _getFieldDict[field]
        except KeyError as e:
            # re-raise the exception with a bit more info
            raise SyntaxError("Cannot get field " + field +
                    " for parameter " + self.name + "\n" + str(e))
        if field == "p_value":
            # return value of parameter
            # Note that IRAF returns the filename for list parameters
            # when p_value is used.  I consider this a bug, and it does
            # not appear to be used by any cl scripts or SPP programs
            # in either IRAF or STSDAS.  It is also in conflict with
            # the IRAF help documentation.  I am making p_value exactly
            # the same as just a simple CL parameter reference.
            return self.get(native=native,prompt=prompt)
        elif field == "p_name": return self.name
        elif field == "p_xtype": return self.type
        elif field == "p_type": return self._getPType()
        elif field == "p_mode": return self.mode
        elif field == "p_prompt": return self.prompt
        elif field == "p_scope": return self.scope
        elif field == "p_default" or field == "p_filename":
            # these all appear to be equivalent -- they just return the
            # current PFilename of the parameter (which is the same as the value
            # for non-list parameters, and is the filename for list parameters)
            return self._getPFilename(native,prompt)
        elif field == "p_maximum":
            if native:
                return self.max
            else:
                return self.toString(self.max)
        elif field == "p_minimum":
            if self.choice is not None:
                if native:
                    return self.choice
                else:
                    schoice = list(map(self.toString, self.choice))
                    return "|" + "|".join(schoice) + "|"
            else:
                if native:
                    return self.min
                else:
                    return self.toString(self.min)
        else:
            # XXX unimplemented fields:
            # p_length: maximum string length in bytes -- what to do with it?
            raise RuntimeError("Program bug in IrafPar._getField()\n" +
                    "Requested field " + field + " for parameter " + self.name)

    def _setField(self, value, field, check=1):
        """Set a parameter field value"""
        try:
            # expand field name using minimum match
            field = _setFieldDict[field]
        except KeyError as e:
            raise SyntaxError("Cannot set field " + field +
                    " for parameter " + self.name + "\n" + str(e))
        if field == "p_prompt":
            self.prompt = irafutils.removeEscapes(irafutils.stripQuotes(value))
        elif field == "p_value":
            self.set(value,check=check)
        elif field == "p_filename":
            # this is only relevant for list parameters (*imcur, *gcur, etc.)
            self.set(value,check=check)
        elif field == "p_scope":
            self.scope = value
        elif field == "p_maximum":
            self.max = self._coerceOneValue(value)
        elif field == "p_minimum":
            if isinstance(value,str) and '|' in value:
                self._setChoice(irafutils.stripQuotes(value))
            else:
                self.min = self._coerceOneValue(value)
        elif field == "p_mode":
            # not doing any type or value checking here -- setting mode is
            # rare, so assume that it is being done correctly
            self.mode = irafutils.stripQuotes(value)
        else:
            raise RuntimeError("Program bug in IrafPar._setField()" +
                    "Requested field " + field + " for parameter " + self.name)

    def _coerceValue(self,value,strict=0):
        """Coerce parameter to appropriate type

        Should accept None or null string.
        """
        return self._coerceOneValue(value,strict)

    def _coerceOneValue(self,value,strict=0):
        """Coerce a scalar parameter to the appropriate type

        Default implementation simply prevents direct use of base class.
        Should accept None or null string.
        """
        raise NotImplementedError("class IrafPar cannot be used directly")


# -----------------------------------------------------
# IRAF array parameter base class
# -----------------------------------------------------

class IrafArrayPar(IrafPar):

    """IRAF array parameter class"""

    def __init__(self,fields,strict=0):
        orig_len = len(fields)
        if orig_len < 3:
            raise SyntaxError("At least 3 fields must be given")
        #
        # all the attributes that are going to get defined
        #
        self.__dict__.update(_IrafPar_attr_dict)
        self.name   = fields[0]
        self.type   = fields[1]
        self.mode   = fields[2]
        self.__dict__['shape'] = None
        #
        # for array parameters, dimensions follow mode field
        # and values come from fields after prompt
        #
        if len(fields)<4 or fields[3] is None:
            raise ValueError("Missing dimension field for array parameter")
        ndim = int(fields[3])
        if len(fields) < 4+2*ndim:
            raise ValueError("Missing array shape fields for array parameter")
        shape = []
        array_size = 1
        for i in range(ndim):
            shape.append(int(fields[4+2*i]))
            array_size = array_size*shape[-1]
        self.shape = tuple(shape)
        nvstart = 7+2*ndim
        fields.extend([""]*(nvstart-len(fields)))
        fields.extend([None]*(nvstart+array_size-len(fields)))
        if len(fields) > nvstart+array_size:
            raise SyntaxError("Too many values for array" +
                    " for parameter " + self.name)
        #
        self.value = [None]*array_size
        self.value = self._coerceValue(fields[nvstart:],strict)
        if fields[nvstart-3] is not None and '|' in fields[nvstart-3]:
            self._setChoice(fields[nvstart-3].strip(),strict)
            if fields[nvstart-2] is not None:
                if orig_len < nvstart:
                    warning("Max value illegal when choice list given" +
                                    " for parameter " + self.name +
                                    " (probably missing comma)",
                                    strict)
                    # try to recover by assuming max string is prompt
                    #XXX risky -- all init values might be off by one
                    fields[nvstart-1] = fields[nvstart-2]
                    fields[nvstart-2] = None
                else:
                    warning("Max value illegal when choice list given" +
                            " for parameter " + self.name, strict)
        else:
            self.min = self._coerceOneValue(fields[nvstart-3],strict)
            self.max = self._coerceOneValue(fields[nvstart-2],strict)
        if fields[nvstart-1] is not None:
            self.prompt = irafutils.removeEscapes(
                                            irafutils.stripQuotes(fields[nvstart-1]))
        else:
            self.prompt = ''
        if self.min not in [None, INDEF] and \
           self.max not in [None, INDEF] and self.max < self.min:
            warning("Maximum " + str(self.max) + " is less than minimum " + \
                    str(self.min) + " for parameter " + self.name,
                    strict)
            self.min, self.max = self.max, self.min
        #
        # check attributes to make sure they are appropriate for
        # this parameter type (e.g. some do not allow choice list
        # or min/max)
        #
        self._checkAttribs(strict)
        #
        # check parameter value to see if it is correct
        #
        try:
            self.checkValue(self.value,strict)
        except ValueError as e:
            warning("Illegal initial value for parameter\n" + str(e),
                    strict, exception=ValueError)
            # Set illegal values to None, just like IRAF
            self.value = None

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def save(self, dolist=0):
        """Return .par format string for this parameter

        If dolist is set, returns fields as a list of strings.  Default
        is to return a single string appropriate for writing to a file.
        """
        quoted = not dolist
        array_size = 1
        for d in self.shape:
            array_size = d*array_size
        ndim = len(self.shape)
        fields = (7+2*ndim+len(self.value))*[""]
        fields[0] = self.name
        fields[1] = self.type
        fields[2] = self.mode
        fields[3] = str(ndim)
        next = 4
        for d in self.shape:
            fields[next] = str(d); next += 1
            fields[next] = '1';    next += 1
        nvstart = 7+2*ndim
        if self.choice is not None:
            schoice = list(map(self.toString, self.choice))
            schoice.insert(0,'')
            schoice.append('')
            fields[nvstart-3] = repr('|'.join(schoice))
        elif self.min not in [None,INDEF]:
            fields[nvstart-3] = self.toString(self.min,quoted=quoted)
        # insert an escaped line break before min field
        if quoted:
            fields[nvstart-3] = '\\\n' + fields[nvstart-3]
        if self.max not in [None,INDEF]:
            fields[nvstart-2] = self.toString(self.max,quoted=quoted)
        if self.prompt:
            if quoted:
                sprompt = repr(self.prompt)
            else:
                sprompt = self.prompt
            # prompt can have embedded newlines (which are printed)
            sprompt = sprompt.replace(r'\012', '\n')
            sprompt = sprompt.replace(r'\n', '\n')
            fields[nvstart-1] = sprompt
        for i in range(len(self.value)):
            fields[nvstart+i] = self.toString(self.value[i],quoted=quoted)
        # insert an escaped line break before value fields
        if dolist:
            return fields
        else:
            fields[nvstart] = '\\\n' + fields[nvstart]
            return ','.join(fields)

    def dpar(self, cl=1):
        """Return dpar-style executable assignment for parameter

        Default is to write CL version of code; if cl parameter is
        false, writes Python executable code instead.  Note that
        dpar doesn't even work for arrays in the CL, so we just use
        Python syntax here.
        """
        sval = list(map(self.toString, self.value, len(self.value)*[1]))
        for i in range(len(sval)):
            if sval[i] == "":
                sval[i] = "None"
        s = "%s = [%s]" % (self.name, ', '.join(sval))
        return s

    def get(self, field=None, index=None, lpar=0, prompt=1, native=0, mode="h"):
        """Return value of this parameter as a string (or in native format
        if native is non-zero.)"""

        if field: return self._getField(field,native=native,prompt=prompt)

        # may prompt for value if prompt flag is set
        #XXX should change _optionalPrompt so we prompt for each element of
        #XXX the array separately?  I think array parameters are
        #XXX not useful as non-hidden params.

        if prompt: self._optionalPrompt(mode)

        if index is not None:
            sumindex = self._sumindex(index)
            try:
                if native:
                    return self.value[sumindex]
                else:
                    return self.toString(self.value[sumindex])
            except IndexError:
                # should never happen
                raise SyntaxError("Illegal index [" + repr(sumindex) +
                        "] for array parameter " + self.name)
        elif native:
            # return object itself for an array because it is
            # indexable, can have values assigned, etc.
            return self
        else:
            # return blank-separated string of values for array
            return str(self)

    def set(self, value, field=None, index=None, check=1):
        """Set value of this parameter from a string or other value.
        Field is optional parameter field (p_prompt, p_minimum, etc.)
        Index is optional array index (zero-based).  Set check=0 to
        assign the value without checking to see if it is within
        the min-max range or in the choice list."""
        if index is not None:
            sumindex = self._sumindex(index)
            try:
                value = self._coerceOneValue(value)
                if check:
                    self.value[sumindex] = self.checkOneValue(value)
                else:
                    self.value[sumindex] = value
                return
            except IndexError:
                # should never happen
                raise SyntaxError("Illegal index [" + repr(sumindex) +
                        "] for array parameter " + self.name)
        if field:
            self._setField(value,field,check=check)
        else:
            if check:
                self.value = self.checkValue(value)
            else:
                self.value = self._coerceValue(value)
            self.setChanged()

    def checkValue(self,value,strict=0):
        """Check and convert a parameter value.

        Raises an exception if the value is not permitted for this
        parameter.  Otherwise returns the value (converted to the
        right type.)
        """
        v = self._coerceValue(value,strict)
        for i in range(len(v)):
            self.checkOneValue(v[i],strict=strict)
        return v

    #--------------------------------------------
    # special methods
    #--------------------------------------------

    # array parameters can be subscripted
    # note subscripts start at zero, unlike CL subscripts
    # that start at one

    def __getitem__(self, index):
        return self.get(index=index,native=1)

    def __setitem__(self, index, value):
        self.set(value, index=index)

    def __str__(self):
        """Return readable description of parameter"""
        # This differs from non-arrays in that it returns a
        # print string with just the values.  That's because
        # the object itself is returned as the native value.
        sv = list(map(str, self.value))
        for i in range(len(sv)):
            if self.value[i] is None:
                sv[i] = "INDEF"
        return ' '.join(sv)

    def __len__(self):
        return len(self.value)

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    def _sumindex(self, index=None):
        """Convert tuple index to 1-D index into value"""
        try:
            ndim = len(index)
        except TypeError:
            # turn index into a 1-tuple
            index = (index,)
            ndim = 1
        if len(self.shape) != ndim:
            raise ValueError("Index to %d-dimensional array %s has too %s dimensions" %
                (len(self.shape), self.name, ["many","few"][len(self.shape) > ndim]))
        sumindex = 0
        for i in range(ndim-1,-1,-1):
            index1 = index[i]
            if index1 < 0 or index1 >= self.shape[i]:
                raise ValueError("Dimension %d index for array %s is out of bounds (value=%d)" %
                    (i+1, self.name, index1))
            sumindex = index1 + sumindex*self.shape[i]
        return sumindex

    def _getPType(self):
        """Get underlying datatype for this parameter (strip off 'a' array params)"""
        return self.type[1:]

    def _coerceValue(self,value,strict=0):
        """Coerce parameter to appropriate type

        Should accept None or null string.  Must be an array.
        """
        try:
            if isinstance(value,str):
                # allow single blank-separated string as input
                value = value.split()
            if len(value) != len(self.value):
                raise IndexError
            v = len(self.value)*[0]
            for i in range(len(v)):
                v[i] = self._coerceOneValue(value[i],strict)
            return v
        except (IndexError, TypeError):
            raise ValueError("Value must be a " + repr(len(self.value)) +
                    "-element array for " + self.name)

    def isLegal(self):
        """Dont call checkValue for arrays"""
        try:
            return self.value is not None
        except ValueError:
            return 0


# -----------------------------------------------------
# IRAF string parameter mixin class
# -----------------------------------------------------

class _StringMixin:

    """IRAF string parameter mixin class"""

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def toString(self, value, quoted=0):
        """Convert a single (non-array) value of the appropriate type for
        this parameter to a string"""
        if value is None:
            return ""
        elif quoted:
            return repr(value)
        else:
            return value

    # slightly modified checkOneValue allows minimum match for
    # choice strings and permits null string as value
    def checkOneValue(self,v,strict=0):
        if v is None or v[:1] == ")":
            return v
        elif self.choice is not None:
            try:
                v = self.choiceDict[v]
            except minmatch.AmbiguousKeyError:
                clist = self.choiceDict.getall(v)
                raise ValueError("Parameter %s: "
                        "ambiguous value `%s', could be %s" %
                        (self.name, str(v), "|".join(clist)))
            except KeyError:
                raise ValueError("Parameter %s: "
                        "value `%s' is not in choice list (%s)" %
                        (self.name, str(v), "|".join(self.choice)))
        elif (self.min is not None and v<self.min):
            raise ValueError("Parameter %s: "
                    "value `%s' is less than minimum `%s'" %
                    (self.name, str(v), str(self.min)))
        elif (self.max is not None and v>self.max):
            raise ValueError("Parameter %s: "
                    "value `%s' is greater than maximum `%s'" %
                    (self.name, str(v), str(self.max)))
        return v

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    def _checkAttribs(self, strict):
        """Check initial attributes to make sure they are legal"""
        if self.min:
            warning("Minimum value not allowed for string-type parameter " +
                    self.name, strict)
        self.min = None
        if self.max:
            if not self.prompt:
                warning("Maximum value not allowed for string-type parameter " +
                                self.name + " (probably missing comma)",
                                strict)
                # try to recover by assuming max string is prompt
                self.prompt = self.max
            else:
                warning("Maximum value not allowed for string-type parameter " +
                        self.name, strict)
        self.max = None
        # If not in strict mode, allow file (f) to act just like string (s).
        # Otherwise choice is also forbidden for file type
        if strict and self.type == "f" and self.choice:
            warning("Illegal choice value for type '" +
                    self.type + "' for parameter " + self.name,
                    strict)
            self.choice = None

    def _setChoiceDict(self):
        """Create min-match dictionary for choice list"""
        # value is full name of choice parameter
        self.choiceDict = minmatch.MinMatchDict()
        for c in self.choice: self.choiceDict.add(c, c)

    def _nullPrompt(self):
        """Returns value to use when answer to prompt is null string"""
        # for string, null string is a legal value
        # keep current default unless it is None
        if self.value is None:
            return ""
        else:
            return self.value

    def _coerceOneValue(self,value,strict=0):
        if value is None:
            return value
        elif isinstance(value,str):
            # strip double quotes and remove escapes before quotes
            return irafutils.removeEscapes(irafutils.stripQuotes(value))
        else:
            return str(value)

# -----------------------------------------------------
# IRAF string parameter class
# -----------------------------------------------------

class IrafParS(_StringMixin, IrafPar):

    """IRAF string parameter class"""
    pass

# -----------------------------------------------------
# IRAF string array parameter class
# -----------------------------------------------------

class IrafParAS(_StringMixin,IrafArrayPar):

    """IRAF string array parameter class"""
    pass

# -----------------------------------------------------
# IRAF boolean parameter mixin class
# -----------------------------------------------------

class _BooleanMixin:

    """IRAF boolean parameter mixin class"""

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def toString(self, value, quoted=0):
        if value in [None, INDEF]:
            return ""
        elif isinstance(value,str):
            # presumably an indirection value ')task.name'
            if quoted:
                return repr(value)
            else:
                return value
        else:
            # must be internal yes, no value
            return str(value)

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    def _checkAttribs(self, strict):
        """Check initial attributes to make sure they are legal"""
        if self.min:
            warning("Minimum value not allowed for boolean-type parameter " +
                    self.name, strict)
            self.min = None
        if self.max:
            if not self.prompt:
                warning("Maximum value not allowed for boolean-type parameter " +
                                self.name + " (probably missing comma)",
                                strict)
                # try to recover by assuming max string is prompt
                self.prompt = self.max
            else:
                warning("Maximum value not allowed for boolean-type parameter " +
                        self.name, strict)
            self.max = None
        if self.choice:
            warning("Choice values not allowed for boolean-type parameter " +
                    self.name, strict)
            self.choice = None

    # accepts special yes, no objects, integer values 0,1 or
    # string 'yes','no' and variants
    # internal value is yes, no, None/INDEF, or indirection string
    def _coerceOneValue(self,value,strict=0):
        if value == INDEF:
            return INDEF
        elif value is None or value == "":
            return None
        elif value in (1, 1.0, yes, "yes", "YES", "y", "Y", True):
            return yes
        elif value in (0, 0.0, no,  "no",  "NO",  "n", "N", False):
            return no
        elif isinstance(value,str):
            v2 = irafutils.stripQuotes(value.strip())
            if v2 == "" or v2 == "INDEF" or \
                    ((not strict) and (v2.upper() == "INDEF")):
                return INDEF
            elif v2[0:1] == ")":
                # assume this is indirection -- just save it as a string
                return v2
        raise ValueError("Parameter %s: illegal boolean value %s or type %s" %
                (self.name, repr(value), str(type(value))))

# -----------------------------------------------------
# IRAF boolean parameter class
# -----------------------------------------------------

class IrafParB(_BooleanMixin,IrafPar):

    """IRAF boolean parameter class"""
    pass

# -----------------------------------------------------
# IRAF boolean array parameter class
# -----------------------------------------------------

class IrafParAB(_BooleanMixin,IrafArrayPar):

    """IRAF boolean array parameter class"""
    pass

# -----------------------------------------------------
# IRAF integer parameter mixin class
# -----------------------------------------------------

class _IntMixin:

    """IRAF integer parameter mixin class"""

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def toString(self, value, quoted=0):
        if value is None:
            return ""
        else:
            return str(value)

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    # coerce value to integer
    def _coerceOneValue(self,value,strict=0):
        if value == INDEF:
            return INDEF
        elif value is None or isinstance(value,int):
            return value
        elif value in ("", "None", "NONE"):
            return None
        elif isinstance(value,float):
            # try converting to integer
            try:
                return int(value)
            except (ValueError, OverflowError):
                pass
        elif isinstance(value,str):
            s2 = irafutils.stripQuotes(value.strip())
            if s2 == "INDEF" or \
              ((not strict) and (s2.upper() == "INDEF")):
                return INDEF
            elif s2[0:1] == ")":
                # assume this is indirection -- just save it as a string
                return s2
            elif s2[-1:] == "x":
                # hexadecimal
                return int(s2[:-1],16)
            elif "." in s2:
                # try interpreting as a float and converting to integer
                try:
                    return int(float(s2))
                except (ValueError, OverflowError):
                    pass
            else:
                try:
                    return int(s2)
                except ValueError:
                    pass
        else:
            # maybe it has an int method
            try:
                return int(value)
            except ValueError:
                pass
        raise ValueError("Parameter %s: illegal integer value %s" %
                (self.name, repr(value)))

# -----------------------------------------------------
# IRAF integer parameter class
# -----------------------------------------------------

class IrafParI(_IntMixin,IrafPar):

    """IRAF integer parameter class"""
    pass

# -----------------------------------------------------
# IRAF integer array parameter class
# -----------------------------------------------------

class IrafParAI(_IntMixin,IrafArrayPar):

    """IRAF integer array parameter class"""
    pass

# -----------------------------------------------------
# Strict integer parameter mixin class
# -----------------------------------------------------

class _StrictIntMixin(_IntMixin):

    """Strict integer parameter mixin class"""

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def toString(self, value, quoted=0):
        return str(value)

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    # coerce value to integer
    def _coerceOneValue(self,value,strict=0):
        if value is None or isinstance(value,int):
            return value
        elif isinstance(value,str):
            s2 = irafutils.stripQuotes(value.strip())
            if s2[-1:] == "x":
                # hexadecimal
                return int(s2[:-1],16)
            elif s2 == '':
                raise ValueError('Parameter '+self.name+ \
                      ': illegal empty integer value')
            else:
                # see if it is a stringified int
                try:
                    return int(s2)
                except ValueError:
                    pass
        # otherwise it is not a strict integer
        raise ValueError("Parameter %s: illegal integer value %s" %
                (self.name, repr(value)))

# -----------------------------------------------------
# Strict integer parameter class
# -----------------------------------------------------

class StrictParI(_StrictIntMixin,IrafPar):

    """Strict integer parameter class"""
    pass


# -----------------------------------------------------
# IRAF real parameter mixin class
# -----------------------------------------------------

_re_d = re.compile(r'[Dd]')
_re_colon = re.compile(r':')

class _RealMixin:

    """IRAF real parameter mixin class"""

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def toString(self, value, quoted=0):
        if value is None:
            return ""
        else:
            return str(value)

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    def _checkAttribs(self, strict):
        """Check initial attributes to make sure they are legal"""
        if self.choice:
            warning("Choice values not allowed for real-type parameter " +
                    self.name, strict)
            self.choice = None

    # coerce value to real
    # using clFloat in Py3 reproduces same numerical values as on Py2
    def _coerceOneValue(self,value,strict=0):
        if value == INDEF:
            return INDEF
        elif value is None or isinstance(value,clFloat):
            return value
        elif isinstance(value,float):
            return clFloat(value)
        elif value in ("", "None", "NONE"):
            return None
        elif isinstance(value, int_types):
            return clFloat(value)
        elif isinstance(value,str):
            s2 = irafutils.stripQuotes(value.strip())
            if s2 == "INDEF" or \
              ((not strict) and (s2.upper() == "INDEF")):
                return INDEF
            elif s2[0:1] == ")":
                # assume this is indirection -- just save it as a string
                return s2
            # allow +dd:mm:ss.s sexagesimal format for floats
            fvalue = 0.0
            vscale = 1.0
            vsign = 1
            i1 = 0
            mm = _re_colon.search(s2)
            if mm is not None:
                if s2[0:1] == "-":
                    i1 = 1
                    vsign = -1
                elif s2[0:1] == "+":
                    i1 = 1
                while mm is not None:
                    i2 = mm.start()
                    fvalue = fvalue + int(s2[i1:i2])/vscale
                    i1 = i2+1
                    vscale = vscale*60.0
                    mm = _re_colon.search(s2,i1)
            # special handling for d exponential notation
            mm = _re_d.search(s2,i1)
            try:
                if mm is None:
                    return clFloat(vsign*(fvalue + float(s2[i1:])/vscale))
                else:
                    return clFloat(vsign*(fvalue + \
                            float(s2[i1:mm.start()]+"E"+s2[mm.end():])/vscale))
            except ValueError:
                pass
        else:
            # maybe it has a float method
            try:
                return clFloat(value)
            except ValueError:
                pass
        raise ValueError("Parameter %s: illegal float value %s" %
                (self.name, repr(value)))


# -----------------------------------------------------
# IRAF real parameter class
# -----------------------------------------------------

class IrafParR(_RealMixin,IrafPar):

    """IRAF real parameter class"""
    pass

# -----------------------------------------------------
# IRAF real array parameter class
# -----------------------------------------------------

class IrafParAR(_RealMixin,IrafArrayPar):

    """IRAF real array parameter class"""
    pass

# -----------------------------------------------------
# Strict real parameter mixin class
# -----------------------------------------------------

class _StrictRealMixin(_RealMixin):

    """Strict real parameter mixin class"""

    #--------------------------------------------
    # public methods
    #--------------------------------------------

    def toString(self, value, quoted=0):
        return str(value)

    #--------------------------------------------
    # private methods
    #--------------------------------------------

    # coerce value to real
    def _coerceOneValue(self,value,strict=0):
        if value is None or isinstance(value,clFloat):
            return value
        elif isinstance(value,float):
            return clFloat(value)
        elif isinstance(value, int_types):
            return clFloat(value)
        elif isinstance(value,str):
            s2 = irafutils.stripQuotes(value.strip())
            if s2 == '':
                raise ValueError('Parameter '+self.name+ \
                      ': illegal empty float value')
            # allow +dd:mm:ss.s sexagesimal format for floats
            fvalue = 0.0
            vscale = 1.0
            vsign = 1
            i1 = 0
            mm = _re_colon.search(s2)
            if mm is not None:
                if s2[0:1] == "-":
                    i1 = 1
                    vsign = -1
                elif s2[0:1] == "+":
                    i1 = 1
                while mm is not None:
                    i2 = mm.start()
                    fvalue = fvalue + int(s2[i1:i2])/vscale
                    i1 = i2+1
                    vscale = vscale*60.0
                    mm = _re_colon.search(s2,i1)
            # special handling for d exponential notation
            mm = _re_d.search(s2,i1)
            try:
                if mm is None:
                    return clFloat(vsign*(fvalue + float(s2[i1:])/vscale))
                else:
                    return clFloat(vsign*(fvalue + \
                            float(s2[i1:mm.start()]+"E"+s2[mm.end():])/vscale))
            except ValueError:
                pass
            # see if it's a stringified float
            try:
                return clFloat(s2)
            except ValueError:
                raise ValueError("Parameter %s: illegal float value %s" %
                                 (self.name, repr(value)))
        # Otherwise it is not a strict float
        raise ValueError("Parameter %s: illegal float value %s" %
                         (self.name, repr(value)))


# -----------------------------------------------------
# Strict real parameter class
# -----------------------------------------------------

class StrictParR(_StrictRealMixin,IrafPar):

    """Strict real parameter class"""
    pass


# -----------------------------------------------------
# Utility routine for parsing choice string
# -----------------------------------------------------

_re_choice = re.compile(r'\|')

def _getChoice(s, strict):
    clist = s.split("|")
    # string is allowed to start and end with "|", so ignore initial
    # and final empty strings
    if not clist[0]: del clist[0]
    if len(clist)>1 and not clist[-1]: del clist[-1]
    return clist



"""eparoption.py: module for defining the various parameter display
   options to be used for the parameter editor task.  The widget that is used
   for entering the parameter value is the variant.  Instances should be
   created using the eparOptionFactory function defined at the end of the
   module.

   Parameter types:
   string  - Entry widget
   *gcur   - NOT IMPLEMENTED AT THIS TIME
   ukey    - NOT IMPLEMENTED AT THIS TIME
   pset    - Action button
   real    - Entry widget
   int     - Entry widget
   boolean - Radiobutton widget
   array real - NOT IMPLEMENTED AT THIS TIME
   array int  - NOT IMPLEMENTED AT THIS TIME

   Enumerated lists - Menubutton/Menu widget

$Id$

M.D. De La Pena, 1999 August 05
"""
# System level modules
import sys
import string

from . import capable

if capable.OF_GRAPHICS:
    from tkinter import *
    from tkinter.filedialog import askdirectory, askopenfilename
else:
    StringVar = None

# Are we using X? (see description of logic in pyraf's wutil.py)
USING_X = True
if sys.platform == 'darwin':
    junk = ",".join(sys.path)
    USING_X = junk.lower().find('/pyobjc') < 0
    del junk

# Constants
MAXLIST  =  15
MAXLINES = 100
XSHIFT   = 110
DSCRPTN_FLAG = ' (***)'


class EparOption:

    """EparOption base class

    Implementation for a specific parameter type must implement
    the makeInputWidget method and must create an attribute `entry'
    with the base widget created.  The entry widget is used for
    focus setting and automatic scrolling.  doScroll is a callback to
    do the scrolling when tab changes focus.
    """

    # Chosen option
    choiceClass = StringVar

    def __init__(self, master, statusBar, paramInfo, defaultParamInfo,
                 doScroll, fieldWidths, defaultsVerb, bg,
                 indent=False, helpCallbackObj=None, mainGuiObj=None):

        # Connect to the information/status Label
        self.status = statusBar

        # Hook to allow scroll when this widget gets focus
        self.doScroll = doScroll
        # Track selection at the last FocusOut event
        self.lastSelection = (0,END)

        # A new Frame is created for each parameter entry
        self.master       = master
        self.bkgColor     = bg
        self.master_frame = Frame(self.master, bg=self.bkgColor)
        self.paramInfo    = paramInfo
        self.defaultParamInfo = defaultParamInfo
        self.defaultsVerb = defaultsVerb
        self.inputWidth   = fieldWidths.get('inputWidth')
        self.valueWidth   = fieldWidths.get('valueWidth')
        self.promptWidth  = fieldWidths.get('promptWidth')

        self.choice = self.choiceClass(self.master_frame)
        self.name = self.paramInfo.name
        self.value = self.paramInfo.get(field = "p_filename", native = 0,
                                        prompt = 0)
        self.previousValue = self.value
        self._editedCallbackObj = None
        self._helpCallbackObj = helpCallbackObj
        self._mainGuiObj = mainGuiObj
        self._lastWidgetEditedVal = None
        self._flagNonDefaultVals = False
        self._flaggedColor = "red"

        # DISABLE any indent for now - not sure why but this causes odd text
        # field sizes in other (unrelated and unindented) parameters...  Maybe
        # because it messes with the total width of the window...
        if 0 and indent:
            self.spacer = Label(self.master_frame, anchor=W, takefocus=0,
                                text="", width=3, bg=self.bkgColor)
            self.spacer.pack(side=LEFT, fill=X, expand=TRUE)

        # Generate the input label
        if self.paramInfo.get(field = "p_mode") == "h":
            self.inputLabel = Label(self.master_frame, anchor = W,
                                    text  = "("+self.getShowName()+")",
                                    width = self.inputWidth, bg=self.bkgColor)
        else:
            self.inputLabel = Label(self.master_frame, anchor = W,
                                    text  = self.getShowName(),
                                    width = self.inputWidth, bg=self.bkgColor)
        self.inputLabel.pack(side = LEFT, fill = X, expand = TRUE)

        # Get the prompt string and determine if special handling is needed
        # Use the prompt/description from the default version, in case they
        # have edited theirs - this is not editable - see ticket #803
        self.prompt = self.defaultParamInfo.get(field="p_prompt", native=0,
                                                prompt=0)

        # Check the prompt to determine how many lines of valid text exist
        lines       = self.prompt.split("\n")
        nlines      = len(lines)
        promptLines = " " + lines[0]
        infoLines   = ""
        blankLineNo = MAXLINES
        if (nlines > 1):
            # Keep all the lines of text before the blank line for the prompt
            for i in range(1, nlines):
                ntokens = lines[i].split()
                if ntokens != []:
                    promptLines = "\n".join([promptLines, lines[i]])
                else:
                    blankLineNo = i
                    break
        self._flagged = False
        if promptLines.endswith(DSCRPTN_FLAG):
            promptLines = promptLines[:-len(DSCRPTN_FLAG)]
            self._flagged = True
        fgColor = "black"
        # turn off this red coloring for the DSCRPTN_FLAG - see #803
#       if self._flagged: fgColor = "red"

        # Generate the prompt label
        self.promptLabel = Label(self.master_frame, anchor=W, fg=fgColor,
                                 text=promptLines, width=self.promptWidth,
                                 bg=self.bkgColor)
        self.promptLabel.pack(side=RIGHT, fill=X, expand=TRUE)

        # Settings for subclasses to override in the makeInputWidget method
        self.isSelectable = True # ie widget has text (num or str) to select

        # Default is none of items on popup menu are activated
        # These can be changed by the makeInputWidget method to customize
        # behavior for each widget.
        self.browserEnabled = DISABLED
        self.clearEnabled = DISABLED
        self.unlearnEnabled = DISABLED
        self.helpEnabled = DISABLED
        if self._helpCallbackObj is not None:
            self.helpEnabled = NORMAL

        # Generate the input widget depending upon the datatype
        self.makeInputWidget()
#       print(self.name, self.__class__) # DBG line

        self.entry.bind('<FocusOut>', self.focusOut, "+")
        self.entry.bind('<FocusIn>', self.focusIn, "+")

        # Trap keys that leave field and validate entry
        self.entry.bind('<Return>', self.entryCheck, "+")
        self.entry.bind('<Shift-Return>', self.entryCheck, "+")
        self.entry.bind('<Tab>', self.entryCheck, "+")
        self.entry.bind('<Shift-Tab>', self.entryCheck, "+")
        self.entry.bind('<Up>', self.entryCheck, "+")
        self.entry.bind('<Down>', self.entryCheck, "+")
        try:
            # special shift-tab binding needed for (some? all?) linux systems
            self.entry.bind('<KeyPress-ISO_Left_Tab>', self.entryCheck, "+")
        except TclError:
            # Ignore exception here, the binding can't be relevant
            # if ISO_Left_Tab is unknown.
            pass

        # Bind the right button to a popup menu of choices
        if USING_X:
            self.entry.bind('<Button-3>', self.popupChoices)
        else:
            self.entry.bind('<Button-2>', self.popupChoices)

        # Pack the parameter entry Frame
        self.master_frame.pack(side=TOP, fill=X, ipady=1)

        # If there is more text associated with this entry, join all the
        # lines of text with the blank line.  This is the "special" text
        # information.
        if (blankLineNo < (nlines - 1)):

            # Put the text after the blank line into its own Frame
            self.master.infoText = Frame(self.master)

            for j in range(blankLineNo + 1, nlines):
                ntokens = lines[j].split()
                if ntokens != []:
                    infoLines = "\n".join([infoLines, lines[j]])
                else:
                    break

            # Assign the informational text to the label and pack
            self.master.infoText.label = Label(self.master.infoText,
                                               text = infoLines,
                                               anchor = W,
                                               bg = self.bkgColor)
            self.master.infoText.label.pack(side = LEFT)
            self.master.infoText.pack(side = TOP, anchor = W)

    def setFlaggedColor(self, colorstr):
        self._flaggedColor = colorstr

    def setIsFlagging(self, isFlagging, redrawImmediately):
        self._flagNonDefaultVals = isFlagging
        if redrawImmediately:
            if self._flagNonDefaultVals:
                curVal = self.choice.get()
            else: # otheriwse we don't care; use None; is ok and faster
                curVal = None
            self.flagThisPar(curVal, True)

    def getShowName(self):
        """ Return the name to be shown in the GUI for this par/option. """
        return self.name

    def extraBindingsForSelectableText(self):
        """ Collect in 1 place the bindings needed for watchTextSelection() """
        # See notes in watchTextSelection
        self.entry.bind('<FocusIn>', self.watchTextSelection, "+")
        self.entry.bind('<ButtonRelease-1>', self.watchTextSelection, "+")
        self.entry.bind('<B1-Motion>', self.watchTextSelection, "+")
        self.entry.bind('<Shift_L>', self.watchTextSelection, "+")
        self.entry.bind('<Left>', self.watchTextSelection, "+")
        self.entry.bind('<Right>', self.watchTextSelection, "+")

    def convertToNative(self, aVal):
        """ The basic type is natively a string. """
        return None if aVal is None else str(aVal)

    def focusOut(self, event=None):
        """Clear selection (if text is selected in this widget)"""
        # do nothing if this isn't a text-enabled widget
        if not self.isSelectable:
            return
        if self.entryCheck(event) is None:
            # Entry value is OK
            # Save the last selection so it can be restored if we
            # come right back to this widget.  Then clear the selection
            # before moving on.
            entry = self.entry
            try:
                if not entry.selection_present():
                    self.lastSelection = None
                else:
                    self.lastSelection = (entry.index(SEL_FIRST),
                                          entry.index(SEL_LAST))
            except AttributeError:
                pass
            if USING_X and sys.platform == 'darwin':
                pass # do nothing here - we need it left selected for cut/paste
            else:
                entry.selection_clear()
        else:
            return "break"

    def watchTextSelection(self, event=None):
        """ Callback used to see if there is a new text selection. In certain
        cases we manually add the text to the clipboard (though on most
        platforms the correct behavior happens automatically). """
        # Note that this isn't perfect - it is a key click behind when
        # selections are made via shift-arrow.  If this becomes important, it
        # can likely be fixed with after().
        if self.entry.selection_present(): # entry must be text entry type
            i1 = self.entry.index(SEL_FIRST)
            i2 = self.entry.index(SEL_LAST)
            if i1 >= 0 and i2 >= 0 and i2 > i1:
                sel = self.entry.get()[i1:i2]
                # Add to clipboard on platforms where necessary.
                print('selected: "'+sel+'"')
#               The following is unneeded if the selected text stays selected
#               when focus is lost or another app is bought to the forground.
#               if sel and USING_X and sys.platform == 'darwin':
#                   clipboard_helper.put(sel, 'PRIMARY')

    def focusIn(self, event=None):
        """Select all text (if applicable) on taking focus"""
        try:
            # doScroll returns false if the call was ignored because the
            # last call also came from this widget.  That avoids unwanted
            # scrolls and text selection when the focus moves in and out
            # of the window.
            if self.doScroll(event):
                self.entry.selection_range(0, END) # select all text in widget
            else:
                # restore selection to what it was on the last FocusOut
                if self.lastSelection:
                    self.entry.selection_range(*self.lastSelection)
        except AttributeError:
            pass

    # Check the validity of the entry
    # If valid, changes the value of the parameter (note that this
    # is a copy, so change is not permanent until save)
    # Parameter change also sets the isChanged flag.
    def entryCheck(self, event=None, repair=True):

        # Make sure the input is legal
        value = self.choice.get()
        try:
            if value != self.previousValue:
                # THIS will likely get into IrafPar's _coerceOneValue()
                self.paramInfo.set(value)
            # fire any applicable triggers, whether value has changed or not
            self.widgetEdited(action='entry')
            return None
        except ValueError as exceptionInfo:
            # Reset the entry to the previous (presumably valid) value
            if repair:
                self.choice.set(self.previousValue)
                self.status.bell()
            errorMsg = str(exceptionInfo)
            if event is not None:
                self.status.config(text = errorMsg)
            # highlight the text again and terminate processing so
            # focus stays in this widget
            self.focusIn(event)
            return "break"


    def widgetEdited(self, event=None, val=None, action='entry', skipDups=True):
        """ A general method for firing any applicable triggers when
            a value has been set.  This is meant to be easily callable from any
            part of this class (or its subclasses), so that it can be called
            as soon as need be (immed. on click?).  This is smart enough to
            be called multiple times, itself handling the removal of any/all
            duplicate successive calls (unless skipDups is False). If val is
            None, it will use the GUI entry's current value via choice.get().
            See teal.py for a description of action.
        """

        # be as lightweight as possible if obj doesn't care about this stuff
        if not self._editedCallbackObj and not self._flagNonDefaultVals:
            return

        # get the current value
        curVal = val # take this first, if it is given
        if curVal is None:
            curVal = self.choice.get()

        # do any flagging
        self.flagThisPar(curVal, False)

        # see if this is a duplicate successive call for the same value
        if skipDups and curVal==self._lastWidgetEditedVal: return

        # pull trigger
        if not self._editedCallbackObj: return
        self._editedCallbackObj.edited(self.paramInfo.scope,
                                       self.paramInfo.name,
                                       self.previousValue, curVal,
                                       action)
        # for our duplicate checker
        self._lastWidgetEditedVal = curVal


    def focus_set(self, event=None):
        """Set focus to input widget"""
        self.entry.focus_set()


    # Generate the the input widget as appropriate to the parameter datatype
    def makeInputWidget(self):
        pass

    def popupChoices(self, event=None):
        """Popup right-click menu of special parameter operations

        Relies on browserEnabled, clearEnabled, unlearnEnabled, helpEnabled
        instance attributes to determine which items are available.
        """
        # don't bother if all items are disabled
        if NORMAL not in (self.browserEnabled, self.clearEnabled,
                          self.unlearnEnabled, self.helpEnabled):
            return

        self.menu = Menu(self.entry, tearoff = 0)
        if self.browserEnabled != DISABLED:
            # Handle file and directory in different functions (tkFileDialog)
            if capable.OF_TKFD_IN_EPAR:
                self.menu.add_command(label   = "File Browser",
                                      state   = self.browserEnabled,
                                      command = self.fileBrowser)
                self.menu.add_command(label   = "Directory Browser",
                                      state   = self.browserEnabled,
                                      command = self.dirBrowser)
            # Handle file and directory in the same function (filedlg)
            else:
                self.menu.add_command(label   = "File/Directory Browser",
                                      state   = self.browserEnabled,
                                      command = self.fileBrowser)
            self.menu.add_separator()
        self.menu.add_command(label   = "Clear",
                              state   = self.clearEnabled,
                              command = self.clearEntry)
        self.menu.add_command(label   = self.defaultsVerb,
                              state   = self.unlearnEnabled,
                              command = self.unlearnValue)
        self.menu.add_command(label   = 'Help',
                              state   = self.helpEnabled,
                              command = self.helpOnParam)

        # Get the current y-coordinate of the Entry
        ycoord = self.entry.winfo_rooty()

        # Get the current x-coordinate of the cursor
        xcoord = self.entry.winfo_pointerx() - XSHIFT

        # Display the Menu as a popup as it is not associated with a Button
        self.menu.tk_popup(xcoord, ycoord)

    def fileBrowser(self):
        """Invoke a tkinter file dialog"""
        if capable.OF_TKFD_IN_EPAR:
           fname = askopenfilename(parent=self.entry, title="Select File")
        else:
            from . import filedlg
            self.fd = filedlg.PersistLoadFileDialog(self.entry,
                              "Select File", "*")
            if self.fd.Show() != 1:
                self.fd.DialogCleanup()
                return
            fname = self.fd.GetFileName()
            self.fd.DialogCleanup()
        if not fname: return # canceled

        self.choice.set(fname)
        # don't select when we go back to widget to reduce risk of
        # accidentally typing over the filename
        self.lastSelection = None

    def dirBrowser(self):
        """Invoke a tkinter directory dialog"""
        if capable.OF_TKFD_IN_EPAR:
            fname = askdirectory(parent=self.entry, title="Select Directory")
        else:
            raise NotImplementedError('Fix popupChoices() logic.')

        if not fname:
            return # canceled

        self.choice.set(fname)
        # don't select when we go back to widget to reduce risk of
        # accidentally typing over the filename
        self.lastSelection = None

    def clearEntry(self):
        """Clear just this Entry"""
        self.entry.delete(0, END)

    def forceValue(self, newVal, noteEdited=False):
        """Force-set a parameter entry to the given value"""
        if newVal is None:
            newVal = ""
        self.choice.set(newVal)
        if noteEdited:
            self.widgetEdited(val=newVal, skipDups=False)
        # WARNING: the value of noteEdited really should be false (default)
        # in most cases because we need the widgetEdited calls to be arranged
        # at one level higher than we are (single param).  We need to allow the
        # caller to first loop over all eparoptions, setting their values
        # without triggering anything, and THEN go through again and run any
        # triggers.

    def unlearnValue(self):
        """Unlearn a parameter value by setting it back to its default"""
        defaultValue = self.defaultParamInfo.get(field = "p_filename",
                            native = 0, prompt = 0)
        self.choice.set(defaultValue)

    def helpOnParam(self):
        """ Try to display help specific to this parameter. """
        if self._helpCallbackObj is not None:
            self._helpCallbackObj.showParamHelp(self.name)

    def setEditedCallbackObj(self, ecbo):
        """ Sets a callback object to be triggred when this option/parameter
            is edited.  The object is expected to have an "edited()" method
            which takes args as shown where it is called in widgetEdited. """
        self._editedCallbackObj = ecbo

    def setActiveState(self, active):
        """ Use this to enable or disable (grey out) a parameter. """
        st = DISABLED
        if active: st = NORMAL
        self.entry.configure(state=st)
        self.inputLabel.configure(state=st)
        self.promptLabel.configure(state=st)

    def flagThisPar(self, currentVal, force):
        """ If this par's value is different from the default value, it is here
        that we flag it somehow as such.  This basic version simply makes the
        surrounding text red (or returns it to normal). May be overridden.
        Leave force at False if you want to allow this mehtod to make smart
        time-saving decisions about when it can skip recoloring because it is
        already the right color. Set force to true if you think we got out
        of sync and need to be fixed. """

        # Get out ASAP if we can
        if (not force) and (not self._flagNonDefaultVals): return

        # handle simple case before comparing values (quick return)
        if force and not self._flagNonDefaultVals:
            self._flagged = False
            self.promptLabel.configure(fg="black")
            return

        # Get/format values to compare
        currentNative = self.convertToNative(currentVal)
        defaultNative = self.convertToNative(self.defaultParamInfo.value)
        # par.value is same as par.get(native=1,prompt=0)

        # flag or unflag as needed
        if currentNative != defaultNative:
            if not self._flagged or force:
                self._flagged = True
                self.promptLabel.configure(fg=self._flaggedColor) # was red
        else: # same as def
            if self._flagged or force:
                self._flagged = False
                self.promptLabel.configure(fg="black")
        # ['red','blue','green','purple','yellow','orange','black']


class EnumEparOption(EparOption):

    def makeInputWidget(self):

        self.unlearnEnabled = NORMAL
        self.isSelectable = False

        # Set the initial value for the button
        self.choice.set(self.value)

        # Need to adjust the value width so the menu button is
        # aligned properly
        if USING_X:
            self.valueWidth = self.valueWidth - 4
        else:
            pass
#           self.valueWidth = self.valueWidth - 0 # looks right on Aqua

        # Generate the button
        self.entry = Menubutton(self.master_frame,
                                 width  = self.valueWidth,
                                 text   = self.choice.get(),      # label
                                 relief = RAISED,
                                 anchor = W,                      # alignment
                                 textvariable = self.choice,      # var to sync
                                 indicatoron  = 1,
                                 takefocus    = 1,
                                 highlightthickness = 1,
                                 activeforeground='black',
                                 fg='black',
                                 bg=self.bkgColor)

        self.entry.menu = Menu(self.entry, tearoff=0,
                               postcommand=self.postcmd,
                               fg = 'black',
                               bg=self.bkgColor)

        # Generate the dictionary of shortcuts using first letter,
        # second if first not available, etc.
        self.shortcuts = {}
        trylist = self.paramInfo.choice
        underline = {}
        charset = string.ascii_lowercase + string.digits
        i = 0
        while trylist:
            trylist2 = []
            for option in trylist:
                # shortcuts dictionary is case-insensitive
                letter = option[i:i+1].lower()
                if letter in self.shortcuts:
                    # will try again with next letter
                    trylist2.append(option)
                elif letter:
                    if letter in charset:
                        self.shortcuts[letter] = option
                        self.shortcuts[letter.upper()] = option
                        underline[option] = i
                    else:
                        # only allow letters, numbers to be shortcuts
                        # keep going in case this is an embedded blank (e.g.)
                        trylist2.append(option)
                else:
                    # no letters left, so no shortcut for this item
                    underline[option] = -1
            trylist = trylist2
            i = i+1

        # Generate the menu options with shortcuts underlined
        for option in self.paramInfo.choice:
            lbl = option
            if lbl=='-': lbl = ' -' # Tk treats '-' as a separator request
            self.entry.menu.add_radiobutton(label       = lbl,
                                            value       = option,
                                            variable    = self.choice,
                                            command     = self.selected,
                                            indicatoron = 0,
                                            underline   = underline[option])

        # set up a pointer from the menubutton back to the menu
        self.entry['menu'] = self.entry.menu

        self.entry.pack(side = LEFT)

        # shortcut keys jump to items
        for letter in self.shortcuts:
            self.entry.bind('<%s>' % letter, self.keypress)

        # Left button sets focus (as well as popping up menu)
        self.entry.bind('<Button-1>', self.focus_set)

    def keypress(self, event):
        """Allow keys typed in widget to select items"""
        try:
            self.choice.set(self.shortcuts[event.keysym])
        except KeyError:
            # key not found (probably a bug, since we intend to catch
            # only events from shortcut keys, but ignore it anyway)
            pass

    def postcmd(self):
        """Make sure proper entry is activated when menu is posted"""
        value = self.choice.get()
        try:
            index = self.paramInfo.choice.index(value)
            self.entry.menu.activate(index)
        except ValueError:
            # initial null value may not be in list
            pass

    def selected(self):
        """They have chosen an enumerated option."""
        self.widgetEdited(action='entry') # kick off any checks that need doin

#   def setActiveState(self, active):
#       [...]
#       for i in range(len(self.paramInfo.choice)):  # this doesn't seem to
#           self.entry.menu.entryconfig(i, state=st) # make the menu text grey
#       [...]



class BooleanEparOption(EparOption):

    def convertToNative(self, aVal):
        """ Convert to native bool; interpret certain strings. """
        if aVal is None:
            return None
        if isinstance(aVal, bool): return aVal
        # otherwise interpret strings
        return str(aVal).lower() in ('1','on','yes','true')

    def makeInputWidget(self):

        self.unlearnEnabled = NORMAL
        self.isSelectable = False

        # Need to buffer the value width so the radio buttons and
        # the adjoining labels are aligned properly
        self.valueWidth = self.valueWidth + 10
        if USING_X:
            self.padWidth = (self.valueWidth // 2) + 5 # looks right
        else:
            self.padWidth = 2 # looks right on Aqua

        # boolean parameters have 3 values: yes, no & undefined
        # Just display two choices (but variable may initially be
        # undefined)
        self.choice.set(self.value)

        self.entry = Frame(self.master_frame,
                           relief    = FLAT,
                           width     = self.valueWidth,
                           takefocus = 1,
                           highlightthickness = 1,
                           bg=self.bkgColor,
                           highlightbackground=self.bkgColor)
        if not USING_X:
            spacerL= Label(self.entry, takefocus=0, text="", width=2,
                           bg=self.bkgColor)
            spacerL.pack(side=LEFT, fill=X, expand=TRUE)
        self.rbyes = Radiobutton(self.entry, text = "Yes",
                                 variable    = self.choice,
                                 value       = "yes",
                                 anchor      = W,
                                 takefocus   = 0,
                                 underline   = 0,
                                 bg = self.bkgColor,
                                 highlightbackground=self.bkgColor)
        self.rbyes.pack(side=LEFT, ipadx=self.padWidth)
        if not USING_X:
            spacerM= Label(self.entry, takefocus=0, text="", width=3,
                           bg=self.bkgColor)
            spacerM.pack(side=LEFT, fill=X, expand=TRUE)
            spacerR = Label(self.entry, takefocus=0, text="", width=2,
                           bg=self.bkgColor)
            spacerR.pack(side=RIGHT, fill=X, expand=TRUE)
        self.rbno  = Radiobutton(self.entry, text = "No",
                                 variable    = self.choice,
                                 value       = "no",
                                 anchor      = W,
                                 takefocus   = 0,
                                 underline   = 0,
                                 bg = self.bkgColor,
                                 highlightbackground=self.bkgColor)
        self.rbno.pack(side = RIGHT, ipadx = self.padWidth)
        self.entry.pack(side = LEFT)

        # keyboard accelerators
        # Y/y sets yes, N/n sets no, space toggles selection
        self.entry.bind('<y>', self.set)
        self.entry.bind('<Y>', self.set)
        self.entry.bind('<n>', self.unset)
        self.entry.bind('<N>', self.unset)
        self.entry.bind('<space>', self.toggle)
        # When variable changes, make sure widget gets focus
        self.choice.trace("w", self.trace)

        # Right-click menu is bound to individual widgets too
        if USING_X:
            self.rbno.bind('<Button-3>', self.popupChoices)
            self.rbyes.bind('<Button-3>', self.popupChoices)
        else:
            self.rbno.bind('<Button-2>', self.popupChoices)
            self.rbyes.bind('<Button-2>', self.popupChoices)
            spacerM.bind('<Button-2>', self.popupChoices)

        # Regular selection - allow immediate trigger/check
        self.rbyes.bind('<Button-1>', self.boolWidgetEditedYes)
        self.rbno.bind('<Button-1>', self.boolWidgetEditedNo)

    def trace(self, *args):
        self.entry.focus_set()

    # Only needed over widgetEdited because the Yes isn't set yet
    def boolWidgetEditedYes(self, event=None): self.widgetEdited(val="yes")

    # Only needed over widgetEdited because the No isn't set yet
    def boolWidgetEditedNo(self, event=None): self.widgetEdited(val="no")

    def set(self, event=None):
        """Set value to Yes"""
        self.rbyes.select()
        self.widgetEdited()

    def unset(self, event=None):
        """Set value to No"""
        self.rbno.select()
        self.widgetEdited()

    def toggle(self, event=None):
        """Toggle value between Yes and No"""
        if self.choice.get() == "yes":
            self.rbno.select()
        else:
            self.rbyes.select()
        self.widgetEdited()

    def setActiveState(self, active):
        st = DISABLED
        if active: st = NORMAL
        self.rbyes.configure(state=st)
        self.rbno.configure(state=st)
        self.inputLabel.configure(state=st)
        self.promptLabel.configure(state=st)


class StringEparOption(EparOption):

    def makeInputWidget(self):

        self.browserEnabled = NORMAL
        self.clearEnabled = NORMAL
        self.unlearnEnabled = NORMAL

        self.choice.set(self.value)
        self.entry = Entry(self.master_frame, width = self.valueWidth,
                     textvariable = self.choice) # , bg=self.bkgColor)
        self.entry.pack(side = LEFT, fill = X, expand = TRUE)
#       self.extraBindingsForSelectableText() # do not use yet


class ActionEparButton(EparOption):

    def getButtonLabel(self):
        return self.value

    def makeInputWidget(self):
#       self.choice.set(self.value)

        self.browserEnabled = DISABLED
        self.clearEnabled = DISABLED
        self.unlearnEnabled = DISABLED
        self.helpEnabled = NORMAL

        # Need to adjust the value width so the button is aligned properly
        if USING_X:
            self.valueWidth = self.valueWidth - 3
        else:
            self.valueWidth = self.valueWidth - 2

        self.isSelectable = False

        # Generate the button
        self.entry = Button(self.master_frame,
                            width   = self.valueWidth,
                            text    = self.getButtonLabel(),
                            relief  = RAISED,
                            background = self.bkgColor,
                            highlightbackground = self.bkgColor,
                            command = self.clicked)
        self.entry.pack(side = LEFT)

    def clicked(self):
        raise NotImplementedError('clicked() must be implemented')

    def unlearnValue(self):
        pass


# widget class that works for numbers and arrays of numbers

class NumberEparOption(EparOption):

    def convertToNative(self, aVal):
        """ Natively as an int. """
        if aVal in (None, '', 'None', 'NONE', 'INDEF'): return None
        return int(aVal)

    def notNull(self, value):
        vsplit = value.split()
        return vsplit.count("INDEF") != len(vsplit)

    def makeInputWidget(self):

        self.browserEnabled = DISABLED
        self.clearEnabled = NORMAL
        self.unlearnEnabled = NORMAL

        # Retain the original parameter value in case of bad entry
        self.previousValue = self.value

        self.choice.set(self.value)
        self.entry = Entry(self.master_frame, width = self.valueWidth,
                           textvariable = self.choice) #, bg=self.bkgColor)
        self.entry.pack(side = LEFT)
#       self.extraBindingsForSelectableText() # do not use yet

    # Check the validity of the entry
    # Note that doing this using the parameter set method automatically
    # checks max, min, special value (INDEF, parameter indirection), etc.
    def entryCheck(self, event = None, repair = True):
        """ Ensure any INDEF entry is uppercase, before base class behavior """
        valupr = self.choice.get().upper()
        if valupr.strip() == 'INDEF':
            self.choice.set(valupr)
        return EparOption.entryCheck(self, event, repair = repair)

# numeric widget class specific to floats

class FloatEparOption(NumberEparOption):

    def convertToNative(self, aVal):
        """ Natively as a float. """
        if aVal in (None, '', 'None', 'NONE', 'INDEF'): return None
        return float(aVal)


# EparOption values for non-string types
_eparOptionDict = { "b": BooleanEparOption,
                    "r": FloatEparOption,
                    "R": FloatEparOption,
                    "d": FloatEparOption,
                    "I": NumberEparOption,
                    "i": NumberEparOption,
                    "z": ActionEparButton,
                    "ar": FloatEparOption,
                    "ai": NumberEparOption,
                  }

def eparOptionFactory(master, statusBar, param, defaultParam,
                      doScroll, fieldWidths,
                      plugIn=None, editedCallbackObj=None,
                      helpCallbackObj=None, mainGuiObj=None,
                      defaultsVerb="Default", bg=None, indent=False,
                      flagging=False, flaggedColor=None):

    """Return EparOption item of appropriate type for the parameter param"""

    # Allow passed-in overrides
    if plugIn is not None:
        eparOption = plugIn

    # If there is an enumerated list, regardless of datatype use EnumEparOption
    elif param.choice is not None:
        eparOption = EnumEparOption

    else:
        # Use String for types not in the dictionary
        eparOption = _eparOptionDict.get(param.type, StringEparOption)

    # Create it
    eo = eparOption(master, statusBar, param, defaultParam, doScroll,
                    fieldWidths, defaultsVerb, bg,
                    indent=indent, helpCallbackObj=helpCallbackObj,
                    mainGuiObj=mainGuiObj)
    eo.setEditedCallbackObj(editedCallbackObj)
    eo.setIsFlagging(flagging, False)
    if flaggedColor:
        eo.setFlaggedColor(flaggedColor)
    return eo



""" Learn basic capabilities here (e.g. can we display graphics?).
This is meant to be fast and light, having no complicated dependencies, so
that any module can fearlessly import this without adverse affects or
performance concerns.

$Id$
"""
import os
import sys

descrip = "basic capabilities file, last edited: 28 Dec 2017"


def is_darwin_and_x():
    """ Convenience function.  Returns True if is an X11-linked Python/tkinter
    build on OSX.  This is intended to be quick and easy without further
    imports.  As a result, this relies on the assumption that on OSX, PyObjC
    is installed (only) in the Framework builds of Python. """
    if not sys.platform == 'darwin':
        return False

    return which_darwin_linkage() == "x11"


def which_darwin_linkage(force_otool_check=False):
    """ Convenience function.  Returns one of ('x11', 'aqua') in answer to the
    question of whether this is an X11-linked Python/tkinter, or a natively
    built (framework, Aqua) one.  This is only for OSX.
    This relies on the assumption that on OSX, PyObjC is installed
    in the Framework builds of Python.  If it doesn't find PyObjC,
    this inspects the actual tkinter library binary via otool.

    One driving requirement here is to try to make the determination quickly
    and quietly without actually importing/loading any GUI libraries.  We
    even want to avoid importing tkinter if we can.
    """

    # sanity check
    if sys.platform != 'darwin':
        raise OSError('Incorrect usage, not on OSX')

    # If not forced to run otool, then make some quick and dirty
    # simple checks/assumptions, which do not add to startup time and do not
    # attempt to initialize any graphics.
    if not force_otool_check:

        # There will (for now) only ever be an aqua-linked Python/tkinter
        # when using Ureka on darwin, so this is an easy short-circuit check.
        if 'UR_DIR' in os.environ:
            return "aqua"

        # There will *usually* be PyObjC modules on sys.path on the natively-
        # linked Python. This is assumed to be always correct on Python 2.x, as
        # of 2012.  This is kludgy but quick and effective.
        sp = ",".join(sys.path)
        sp = sp.lower().strip(',')
        if '/pyobjc' in sp or 'pyobjc,' in sp or 'pyobjc/' in sp or sp.endswith('pyobjc'):
            return "aqua"

        # Try one more thing - look for the physical PyObjC install dir under site-packages
        # The assumption above using sys.path does not seem to be correct as of the
        # combination of Python2.7.9/PyObjC3.0.4/2015.
        sitepacksloc = os.path.split(os.__file__)[0]+'/site-packages/objc'
        if os.path.exists(sitepacksloc):
            return "aqua"

        # OK, no trace of PyObjC found - need to fall through to the forced otool check.

    # Use otool shell command
    import tkinter as TKNTR
    import subprocess  # nosec
    try:
        tk_dyn_lib = TKNTR._tkinter.__file__
    except AttributeError: # happens on Ureka
        if 'UR_DIR' in os.environ:
            return 'aqua'
        else:
            return 'unknown'
    libs = subprocess.check_output(('/usr/bin/otool', '-L', tk_dyn_lib)).decode('ascii')  # nosec
    if libs.find('/libX11.') >= 0:
        return "x11"
    else:
        return "aqua"


def get_dc_owner(raises, mask_if_self):
    """ Convenience function to return owner of /dev/console.
    If raises is True, this raises an exception on any error.
    If not, it returns any error string as the owner name.
    If owner is self, and if mask_if_self, returns "<self>"."""
    try:
        from pwd import getpwuid
        owner_uid = os.stat('/dev/console').st_uid
        self_uid  = os.getuid()
        if mask_if_self and owner_uid == self_uid:
            return "<self>"
        owner_name = getpwuid(owner_uid).pw_name
        return owner_name
    except Exception as e:
        if raises:
            raise e
        else:
            return str(e)


OF_GRAPHICS = True

if 'PYRAF_NO_DISPLAY' in os.environ or 'PYTOOLS_NO_DISPLAY' in os.environ:
    OF_GRAPHICS = False

if OF_GRAPHICS and sys.platform == 'darwin':
    #
    # On OSX, there is an AppKit error where Python itself will abort if
    # tkinter operations (e.g. tkinter._test() ...) are attempted when running
    # from a remote terminal.  In these situations, it is not even safe to put
    # the code in a try/except block, since the AppKit error seems to happen
    # *asynchronously* within ObjectiveC code.  See PyRAF ticket #149.
    #
    # SO, let's try a quick simple test here (only on OSX) to find out if we
    # are the "console user".  If we are not, then we don't even want to attempt
    # any windows/graphics calls.  See "console user" here:
    #     http://developer.apple.com/library/mac/#technotes/tn2083/_index.html
    # If we are the console user, we own /dev/console and can read from it.
    # When no one is logged in, /dev/console is owned by "root". When user "bob"
    # is logged in locally/physically, /dev/console is owned by "bob".
    # However, if "bob" restarts the X server while logged in, /dev/console
    # may be owned by "sysadmin" - so we check for that.
    #
    if 'PYRAF_YES_DISPLAY' not in os.environ:
        # the use of PYRAF_YES_DISPLAY is a temporary override while we
        # debug why a user might have no read-acces to /dev/console
        dc_owner = get_dc_owner(False, False)
        OF_GRAPHICS = dc_owner == 'sysadmin' or os.access("/dev/console", os.R_OK)

    # Add a double-check for remote X11 users.  We *think* this is a smaller
    # set of cases, so we do it last minute here:
    if not OF_GRAPHICS:
        # On OSX, but logged in remotely. Normally (with native build) this
        # means there are no graphics.  But, what if they're calling an
        # X11-linked Python?  Then we should allow graphics to be attempted.
        OF_GRAPHICS = is_darwin_and_x()

        # OF_GRAPHICS will be True here in only two cases (2nd should be rare):
        #    An OSX Python build linked with X11, or
        #    An OSX Python build linked natively where PyObjC was left out

# After all that, we may have decided that we want graphics.  Now
# that we know it is ok to try to import tkinter, we can test if it
# is there.  If it is not, we are not capable of graphics.
if OF_GRAPHICS :
    try :
        import tkinter as TKNTR
    except ImportError :
        TKINTER_IMPORT_FAILED = 1
        OF_GRAPHICS = False

# Using tkFileDialog from PyRAF (and maybe in straight TEAL) is crashing python
# itself on OSX only.  Allow on Linux.  Mac: use this until PyRAF #171 fixed.
OF_TKFD_IN_EPAR = True
if sys.platform == 'darwin' and OF_GRAPHICS and \
   not is_darwin_and_x(): # if framework ver
    OF_TKFD_IN_EPAR = 'TEAL_TRY_TKFD' in list(os.environ.keys())



"""These were tests under cli in pandokia."""
import io
import math
from contextlib import contextmanager, redirect_stderr
import pytest

from .utils import HAS_IRAF

from .. import iraf
from .. import scanf
from .. import pyrafglobals


@contextmanager
def use_ecl(flag):
    """Temporary enable/disable usage of ECL"""
    # This is a quite dirty hack that is adjusted to do just enough to
    # let the tests here pass. It is not meant to be used outside.
    # Re-initialization of IRAF takes more that what is done here.
    if flag != pyrafglobals._use_ecl:
        old_ecl = pyrafglobals._use_ecl
        pyrafglobals._use_ecl = flag
        from .. import iraffunctions
        iraffunctions._pkgs.clear()
        iraffunctions._tasks.clear()
        iraf.Init(doprint=False, hush=True)
    yield
    if flag != pyrafglobals._use_ecl:
        pyrafglobals._use_ecl = old_ecl
        iraffunctions._pkgs.clear()
        iraffunctions._tasks.clear()
        iraf.Init(doprint=False, hush=True)


@pytest.mark.parametrize('ecl_flag', [False, True])
def test_division_task(ecl_flag):
    # Show how a .cl script would be run
    with use_ecl(ecl_flag):
        stdout = io.StringIO()
        iraf.task(xyz='print "e: " (9/5)\nprint "f: " (9/5.)\n'
                  'print "g: " (9//5)\nprint "h: " (9//5.)',
                  IsCmdString=True)
        iraf.xyz(StdoutAppend=stdout)
        assert stdout.getvalue() == "e: 1\nf: 1.8\ng: 95\nh: 95.0\n"


@pytest.mark.parametrize('arg,expected', [
    ('9/5', '1'),
    ('9/5.', '1.8'),
    ('9//5', '95'),     # string concatenation
    ('9//5.', '95.0'),  # string concatenation
])
@pytest.mark.parametrize('ecl_flag', [False, True])
def test_division(arg, expected, ecl_flag):
    with use_ecl(ecl_flag):
        stdout = io.StringIO()
        iraf.clExecute(f'print "i: " ({arg})', StdoutAppend=stdout)
        assert stdout.getvalue().strip() == f'i: {expected}'


@pytest.mark.parametrize('arg,fmt,expected', [
    ("seven 6 4.0 -7", "%s %d %g %d", ('seven', 6, 4.0, -7)), # aliveness
    ("seven", "%d", None),
    ("seven", "%c%3c%99c", ('s', 'eve', 'n')),
    ("seven", "%[sev]", ('seve', )),
    ("0xabc90", "%x", (703632, )),
])  
def test_scanf(arg, fmt, expected):
    """A basic unit test that sscanf was built/imported correctly and
    can run.
    """
    l = scanf.scanf(fmt, arg)
    assert l == expected


@pytest.mark.skipif(not HAS_IRAF, reason='Need IRAF to run')
@pytest.mark.parametrize('arg,expected', [
    ("pw", "user.pwd"),
    ("lang", "clpackage.language"),
    ("st", "plot.stdplot plot.stdgraph user.strings imcoords.starfind"),
    ("std", "plot.stdplot plot.stdgraph"),
    ("stdg", "plot.stdgraph"),
    ("stdpl", "plot.stdplot"),
    ("star", "imcoords.starfind"),
    ("star man", "imcoords.starfind user.man"),
    ("vi", "tv.vimexam user.vi"),
    ("noao", "clpackage.noao"),
    ("impl", "plot.implot"),
    ("ls", "user.ls"),
    ("surf", "plot.surface noao.surfphot utilities.surfit"),
    ("surf man", "plot.surface noao.surfphot utilities.surfit user.man"),
    ("smart man", "user.man"),
    ("img", "imutil.imgets images.imgeom"),
    ("prot", "system.protect clpackage.proto"),
    ("pro", "plot.prows plot.prow system.protect clpackage.proto"),
    ("prow", "plot.prows plot.prow"),
    ("prows", "plot.prows"),
    ("prowss", None),
    ("dis", "tv.display system.diskspace"),
    ("no", "noao.nobsolete clpackage.noao"),
])
def test_whereis(arg, expected):
    iraf.plot(_doprint=0)
    iraf.images(_doprint=0)

    stdout = io.StringIO()
    args = arg.split(" ")  # convert to a list
    kw = {'StderrAppend': stdout}
    iraf.whereis(*args, **kw)  # catches stdout+err

    # Check that we get one line per argument
    assert len(stdout.getvalue().splitlines()) == len(args)

    # Check that each argument appears in the coresponding line
    for res, a in zip(stdout.getvalue().splitlines(), args):
        assert a in res

    # If task not found
    if expected is None:
        assert f"{arg}: task not found." in stdout.getvalue()
    else:
        # Check that all expected found places are returned
        for f in expected.split():
            assert f in stdout.getvalue()


@pytest.mark.skipif(not HAS_IRAF, reason='Need IRAF to run')
@pytest.mark.parametrize('arg, expected', [
    ("pw", "user"),
    ("lang", "clpackage"),
    ("stdg", "plot"),
    ("stdp", "plot"),
    ("star", "imcoords"),
    ("star man", "imcoords\nuser"),
    ("vi", "user"),
    ("noao", "clpackage"),
    ("impl", "plot"),
    ("ls", "user"),
    ("surf", "\"Task `surf' is ambiguous, could be "
             "utilities.surfit, noao.surfphot, plot.surface\""),
    ("surface", "plot"),
    ("img", "\"Task `img' is ambiguous, could be "
            "images.imgeom, imutil.imgets\""),
    ("pro", "\"Task `pro' is ambiguous, could be "
            "clpackage.proto, system.protect, plot.prow, ...\""),
    ("prot", "\"Task `prot' is ambiguous, could be "
             "clpackage.proto, system.protect\""),
    ("prow", "plot"),
    ("prows", "plot"),
    ("prowss", "prowss: task not found."),
    ("dis", "\"Task `dis' is ambiguous, could be "
            "system.diskspace, tv.display\""),
])
def test_which(arg, expected):
    iraf.plot(_doprint=0)
    iraf.images(_doprint=0)

    # To Test: normal case, disambiguation, ambiguous, not found, multiple
    #          inputs for a single command

    stdout = io.StringIO()
    args = arg.split(" ")  # convert to a list
    kw = {"StderrAppend": stdout}
    iraf.which(*args, **kw)  # catches stdout+err
    actual_lines = stdout.getvalue().strip().splitlines()
    expected_lines = expected.splitlines()
    assert len(actual_lines) == len(expected_lines)
    for arg, actual_line, expected_line in zip(args, actual_lines,
                                               expected_lines):
        if 'ambiguous' in actual_line:
            if 'ambiguous' in expected_line:
                assert actual_line.startswith(expected_line.rstrip('"'))
            else:
                assert f'{expected_line}.{arg}' in actual_line
        else:
            assert actual_line == expected_line


def test_parse_cl_array_subscripts():
    # Check that square brackets are parsed correctly
    # https://github.com/iraf-community/pyraf/issues/115
    stdout = io.StringIO()
    iraf.task(xyz='char cenwavvalue[4]\n'
              'cenwavvalue[1] = "580"\n'
              'print(cenwavvalue[1])',
              IsCmdString=True)
    iraf.xyz(StdoutAppend=stdout)
    assert stdout.getvalue() == "580\n"


@pytest.mark.parametrize('call, expected', [
    ('acos(0.67)', math.acos(0.67)),
    ('asin(0.67)', math.asin(0.67)),
    ('atan2(2.,3.)', math.atan2(2.,3.)),
    ('cos(1.2)', math.cos(1.2)),
    ('dacos(0.67)', math.degrees(math.acos(0.67))),
    ('dasin(0.67)', math.degrees(math.asin(0.67))),
    ('datan2(2.,3.)', math.degrees(math.atan2(2.,3.))),
    ('dcos(12.)', math.cos(math.radians(12.))),
    ('deg(1.2)', math.degrees(1.2)),
    ('dsin(12.)', math.sin(math.radians(12.))),
    ('dtan(12.)', math.tan(math.radians(12.))),
    ('exp(1.2)', math.exp(1.2)),
    ('frac(1.2)', 1.2 % 1),
    ('hypot(2.,3.)', math.hypot(2.,3.)),
    ('int(-1.2)', int(-1.2)),
    ('isindef(0)', False),
    ('isindef(INDEF)', True),
    ('log(1.2)', math.log(1.2)),
    ('log10(1.2)', math.log10(1.2)),
    ('max(2,3,4,2)', max(2,3,4,2)),
    ('min(2,3,4,2)', min(2,3,4,2)),
    ('mod(123, 7)', 123 % 7),
    ('nint(1.5)', 2),
    ('nint(2.5)', 3),
    ('rad(12)', math.radians(12)),
    ('radix(123, 10)', '123'),
    ('radix(123, 16)', '7B'),
    ('radix(-123, 10)', '4294967173'),
    ('radix(-123, 16)', 'FFFFFF85'),
    ('real("1.23")', 1.23),
    ('sign(-1.2)', -1),
    ('sign(1.2)', 1),
    ('sin(1.2)', math.sin(1.2)),
    ('sqrt(1.2)', math.sqrt(1.2)),
    ('stridx("fd", "abcdefg")', 4),
    ('strldx("fd", "abcdefg")', 6),
    ('strlen("abcdefg")', 7),
    ('strlstr("def", "abcdefg")', 4),
    ('strlwr("aBcDeF")', "abcdef"),
    ('strstr("def", "abcdefg")', 4),
    ('strupr("aBcDeF")', "ABCDEF"),
    ('substr("abcdefg", 2, 5)', "bcde"),
    ('tan(1.2)', math.tan(1.2)),
    ('trim("--abcdefg---", "-")', "abcdefg"),
    ('triml("--abcdefg---", "-")', "abcdefg---"),
    ('trimr("--abcdefg---", "-")', "--abcdefg"),
    ])
def test_intrinsic_functions(call, expected):
    with redirect_stderr(io.StringIO()):  # silence task redefinition warnings
        iraf.task(xyz=f'print({call})', IsCmdString=True)
    stdout = io.StringIO()
    iraf.xyz(Stdout=stdout)
    if isinstance(expected, float):
        assert float(stdout.getvalue().strip()) == pytest.approx(expected, 1e-8)
    elif isinstance(expected, int):
        assert int(stdout.getvalue().strip()) == expected
    else:
        assert stdout.getvalue().strip() == expected


@pytest.mark.parametrize('ecl_flag', [False, True])
@pytest.mark.parametrize('encoding', ['utf-8', 'iso-8859-1'])
@pytest.mark.parametrize('code,expected', [
    ('print AA #  Ångström\n', 'AA\n'),
    ('print("test") | cat', 'test\n'),
])
def test_clfile(tmpdir, ecl_flag, encoding, code, expected):
    # Check proper reading of CL files with different encodings
    fname = tmpdir / 'cltestfile.cl'
    with fname.open("w", encoding=encoding) as fp:
        fp.write(code)

    stdout = io.StringIO()

    with use_ecl(ecl_flag):
        iraf.task(xyz=str(fname))
        iraf.xyz(StdoutAppend=stdout)
        assert stdout.getvalue() == expected


@pytest.mark.skipif(not HAS_IRAF, reason='Need IRAF to run')
@pytest.mark.parametrize('ecl_flag', [False, True])
@pytest.mark.parametrize('code', [
    'cat {datafile}',                     # foreign
    'type {datafile} map_cc=no',          # internal
    'concatenate {datafile} out_type=t',  # system
])
@pytest.mark.parametrize('data', [
    b'test for plain ASCII',
    'Unicode Ångström'.encode('utf-8'),
    b'\x8fq\x96\x7f\xe6\xfd\xf8\x12\xb03{U\x81\x11\x014',  # some random data
])
@pytest.mark.parametrize('redir_stdin', [False, True])
@pytest.mark.parametrize('redir_stdout', [False, True])
def test_redir_data(tmpdir, ecl_flag, code, data, redir_stdin, redir_stdout):
    datafile = tmpdir / 'cltestdata.dat'
    with datafile.open('wb') as fp:
        fp.write(data)
    if redir_stdin:
        code = code.format(datafile='') + f' < {str(datafile)}'
    else:
        code = code.format(datafile=str(datafile))

    if redir_stdout:
        outfile = tmpdir / 'cltestoutput.dat'
        code += f' > {str(outfile)}'

    with use_ecl(ecl_flag):
        iraf.task(xyz=code, IsCmdString=True)
        stdout = io.TextIOWrapper(io.BytesIO())  # Emulate a "real" stdout
        iraf.xyz(Stdout=stdout)

    if redir_stdout:
        with open(outfile, 'rb') as fp:
            assert fp.read() == data
    else:
        assert stdout.buffer.getvalue() == data


@pytest.mark.skipif(not HAS_IRAF, reason='Need IRAF to run')
def test_binary_stdout(tmpdir):
    # Test writing binary data to STDOUT, one of several issues mentioned in
    # https://github.com/iraf-community/pyraf/issues/117
    outfile = str(tmpdir / 'testout.gki')
    expected = [
        f"METAFILE '{outfile}':",
        '[1] (2855 words) The SINC Function',
        '[2] (5701 words) .2',
        '[3] (2525 words) Line 250 of dev$pix[200:300,*]',
        '[4] (7637 words) Log Scaling',
        '[5] (97781 words) NOAO/IRAF V2.3 tody@lyra Fri 23:30:27 08-Aug-86',
        '[6] (2501 words) The Sinc Function'
    ]
    iraf.gkiextract('dev$vdm.gki', '2-7', iraf.yes, verify=False,
                    Stdout=outfile)
    stdout = io.StringIO()
    iraf.gkidir(outfile, Stdout=stdout)
    # Require gkidir output to match list above, ignoring spacing:
    assert [' '.join(line.split())
            for line in stdout.getvalue().splitlines() if line] == expected

def test_real_parameter_str_precision():
    # Check that PyRAF uses the same precision for real parameters on Python 3
    # as on 2 (see https://github.com/iraf-community/pyraf/issues/127)
    iraf.task(
        print_real='''procedure print_real(value)
                      real value
                      begin
                          print(value)
                      end''',
        IsCmdString=True
    )
    stdout = io.StringIO()
    iraf.print_real(1.2345678901234567, Stdout=stdout)
    assert stdout.getvalue() == "1.23456789012\n"
    stdout = io.StringIO()
    iraf.print_real(1000000000000000.0, Stdout=stdout)
    assert stdout.getvalue() == "1e+15\n"



####
#       Class FileDialog
#
#       Purpose
#       -------
#
#       FileDialog's are widgets that allow one to select file names by
#       clicking on file names, directory names, filters, etc.
#
#       Standard Usage
#       --------------
#
#       F = FileDialog(widget, some_title, some_filter)
#       if F.Show() != 1:
#               F.DialogCleanup()
#       return
#               file_name = F.GetFileName()
#               F.DialogCleanup()
####
"""
$Id$
"""
import os
from . import capable

from subprocess import getoutput  # nosec

if capable.OF_GRAPHICS:
    import tkinter as TKNTR
    from . import alert
    from .dialog import *  # noqa
else:
    ModalDialog = object


class FileDialog(ModalDialog):

    # constructor

    lastWrtPrtChoice = None

    def __init__(self, widget, title, filter="*", initWProtState=None):
        """ Supply parent widget, title, filter, and initWProtState (True or
        False).  Set initWProtState to None to hide the write-protect
        check-box. """
        self.widget = widget
        self.filter = filter.strip()
        self.orig_dir = os.getcwd()
        self.cwd = os.getcwd()       # the logical current working directory
        self.showChmod = initWProtState is not None
        # normally we use persistence for lastWrtPrtChoice; use this 1st time
        if FileDialog.lastWrtPrtChoice is None:
            FileDialog.lastWrtPrtChoice = initWProtState
        # Allow a start-directory as part of the given filter
        if self.filter.find(os.sep) >= 0:
            self.cwd = os.path.dirname(self.filter)
            self.filter = os.path.basename(self.filter) # do this second!
        # main Dialog code
        Dialog.__init__(self, widget)

    # setup routine called back from Dialog

    def SetupDialog(self):

        # directory label

        self.dirFrame = Frame(self.top)
        self.dirFrame['relief'] = 'raised'
        self.dirFrame['bd']      = '2'
        self.dirFrame.pack({'expand':'no', 'side':'top', 'fill':'both'})
        self.dirLabel = Label(self.dirFrame)
        self.dirLabel["text"] = "Directory:"
        self.dirLabel.pack({'expand':'no', 'side':'left', 'fill':'none'})

        # editable filter

        self.filterFrame = Frame(self.top)
        self.filterFrame['relief'] = 'raised'
        self.filterFrame['bd']   = '2'
        self.filterFrame.pack({'expand':'no', 'side':'top', 'fill':'both'})
        self.filterLabel = Label(self.filterFrame)
        self.filterLabel["text"] = "Filter:"
        self.filterLabel.pack({'expand':'no', 'side':'left', 'fill':'none'})
        self.filterEntry = Entry(self.filterFrame)
        self.filterEntry.bind('<Return>', self.FilterReturnKey)
        self.filterEntry["width"]  = "40"
        self.filterEntry["relief"] = "ridge"
        self.filterEntry.pack({'expand':'yes', 'side':'right', 'fill':'x'})
        self.filterEntry.insert(0, self.filter)

        # the directory and file listboxes

        self.listBoxFrame = Frame(self.top)
        self.listBoxFrame['relief'] = 'raised'
        self.listBoxFrame['bd']  = '2'
        self.listBoxFrame.pack({'expand':'yes', 'side' :'top',
                'pady' :'2', 'padx': '0', 'fill' :'both'})
        self.CreateDirListBox()
        self.CreateFileListBox()
        self.UpdateListBoxes()

        # write-protect option

        junk = FileDialog.lastWrtPrtChoice
        if junk is None:
            junk = 0
        self.wpVar = IntVar(value=junk) # use class attr
        if self.showChmod:
            self.writeProtFrame = Frame(self.top)
            self.writeProtFrame['relief'] = 'raised'
            self.writeProtFrame['bd'] = '2'
            self.writeProtFrame.pack({'expand':'no','side':'top','fill':'both'})
            self.wpButton = Checkbutton(self.writeProtFrame,
                                        text="Write-protect after save",
                                        command=self.wrtPrtClick,
                                        var=self.wpVar)
            self.wpButton.pack({'expand':'no', 'side':'left'})

        # editable filename

        self.fileNameFrame = Frame(self.top)
        self.fileNameFrame.pack({'expand':'no', 'side':'top', 'fill':'both'})
        self.fileNameFrame['relief'] = 'raised'
        self.fileNameFrame['bd']         = '2'
        self.fileNameLabel = Label(self.fileNameFrame)
        self.fileNameLabel["text"] = "File:"
        self.fileNameLabel.pack({'expand':'no', 'side':'left', 'fill':'none'})
        self.fileNameEntry = Entry(self.fileNameFrame)
        self.fileNameEntry["width"]  = "40"
        self.fileNameEntry["relief"] = "ridge"
        self.fileNameEntry.pack({'expand':'yes', 'side':'right', 'fill':'x',
                                 'pady': '2'})
        self.fileNameEntry.bind('<Return>', self.FileNameReturnKey)

        # buttons - ok, filter, cancel

        self.buttonFrame = Frame(self.top)
        self.buttonFrame['relief'] = 'raised'
        self.buttonFrame['bd']   = '2'
        self.buttonFrame.pack({'expand':'no', 'side':'top', 'fill':'x'})
        self.okButton = Button(self.buttonFrame)
        self.okButton["text"]     = "OK"
        self.okButton["command"]   = self.OkPressed
        self.okButton["width"] = 8
        self.okButton.pack({'expand':'yes', 'pady':'2', 'side':'left'})
        self.filterButton = Button(self.buttonFrame)
        self.filterButton["text"]         = "Filter"
        self.filterButton["command"]   = self.FilterPressed
        self.filterButton["width"] = 8
        self.filterButton.pack({'expand':'yes', 'pady':'2', 'side':'left'})
        button = Button(self.buttonFrame)
        button["text"] = "Cancel"
        button["command"] = self.CancelPressed
        button["width"] = 8
        button.pack({'expand':'yes', 'pady':'2', 'side':'left'})

    # create the directory list box

    def CreateDirListBox(self):
        frame = Frame(self.listBoxFrame)
        frame.pack({'expand':'yes', 'side' :'left', 'pady' :'1',
                'fill' :'both'})
        frame['relief'] = 'raised'
        frame['bd']      = '2'
        filesFrame = Frame(frame)
        filesFrame['relief'] = 'flat'
        filesFrame['bd']         = '2'
        filesFrame.pack({'side':'top', 'expand':'no', 'fill':'x'})
        label = Label(filesFrame)
        label['text'] = 'Directories:'
        label.pack({'side':'left', 'expand':'yes', 'anchor':'w',
                'fill':'none'})
        scrollBar = Scrollbar(frame, {'orient':'vertical'})
        scrollBar.pack({'expand':'no', 'side':'right', 'fill':'y'})
        self.dirLb = Listbox(frame, {'yscroll':scrollBar.set})
        self.dirLb.pack({'expand':'yes', 'side' :'top', 'pady' :'1',
                'fill' :'both'})
        self.dirLb.bind('<1>', self.DoSelection)
        self.dirLb.bind('<Double-Button-1>', self.DoDoubleClickDir)
        scrollBar['command'] = self.dirLb.yview

    # create the files list box

    def CreateFileListBox(self):
        frame = Frame(self.listBoxFrame)
        frame['relief'] = 'raised'
        frame['bd']      = '2'
        frame.pack({'expand':'yes', 'side' :'left', 'pady' :'1', 'padx' :'1',
                'fill' :'both'})
        filesFrame = Frame(frame)
        filesFrame['relief'] = 'flat'
        filesFrame['bd']         = '2'
        filesFrame.pack({'side':'top', 'expand':'no', 'fill':'x'})
        label = Label(filesFrame)
        label['text'] = 'Files:'
        label.pack({'side':'left', 'expand':'yes', 'anchor':'w',
                'fill':'none'})
        scrollBar = Scrollbar(frame, {'orient':'vertical'})
        scrollBar.pack({'side':'right', 'fill':'y'})
        self.fileLb = Listbox(frame, {'yscroll':scrollBar.set})
        self.fileLb.pack({'expand':'yes', 'side' :'top', 'pady' :'0',
                'fill' :'both'})
        self.fileLb.bind('<1>', self.DoSelection)
        self.fileLb.bind('<Double-Button-1>', self.DoDoubleClickFile)
        scrollBar['command'] = self.fileLb.yview

    # update the listboxes and directory label after a change of directory

    def UpdateListBoxes(self):
        cwd = self.cwd
        self.fileLb.delete(0, self.fileLb.size())
        filter = self.filterEntry.get()
        # '*' will list recursively, we don't want that.
        if filter == '*':
            filter = ''
        cmd = "/bin/ls " + os.path.join(cwd, filter)
        cmdOutput = getoutput(cmd)  # nosec
        files = cmdOutput.split("\n")
        files.sort()
        for i in range(len(files)):
            if os.path.isfile(os.path.join(cwd, files[i])):
                self.fileLb.insert('end', os.path.basename(files[i]))
        self.dirLb.delete(0, self.dirLb.size())
        files = os.listdir(cwd)
        if cwd != '/':
            files.append('..')
        files.sort()
        for i in range(len(files)):
            if os.path.isdir(os.path.join(cwd, files[i])):
                self.dirLb.insert('end', files[i])
        self.dirLabel['text'] = "Directory:" + self.cwd_print()

    # selection handlers

    def DoSelection(self, event):
        lb = event.widget
        field = self.fileNameEntry
        field.delete(0, AtEnd())
        field.insert(0, os.path.join(self.cwd_print(), lb.get(lb.nearest(event.y))))
        if TKNTR.TkVersion >= 4.0:
            lb.select_clear(0, "end")
            lb.select_anchor(lb.nearest(event.y))
        else:
            lb.select_clear()
            lb.select_from(lb.nearest(event.y))

    def DoDoubleClickDir(self, event):
        lb = event.widget
        self.cwd = os.path.join(self.cwd, lb.get(lb.nearest(event.y)))
        self.UpdateListBoxes()

    def DoDoubleClickFile(self, event):
        self.OkPressed()

    def OkPressed(self):
        self.TerminateDialog(1)

    def wrtPrtClick(self):
        FileDialog.lastWrtPrtChoice = self.wpVar.get() # update class attr

    def FileNameReturnKey(self, event):
        # if its a relative path then include the cwd in the name
        name = self.fileNameEntry.get().strip()
        if not os.path.isabs(os.path.expanduser(name)):
            self.fileNameEntry.delete(0, 'end')
            self.fileNameEntry.insert(0, os.path.join(self.cwd_print(), name))
        self.okButton.flash()
        self.OkPressed()

    def FilterReturnKey(self, event):
        filter = self.filterEntry.get().strip()
        self.filterEntry.delete(0, 'end')
        self.filterEntry.insert(0, filter)
        self.filterButton.flash()
        self.UpdateListBoxes()

    def FilterPressed(self):
        self.UpdateListBoxes()

    def CancelPressed(self):
        self.TerminateDialog(0)

    def GetFileName(self):
        return self.fileNameEntry.get()

    def GetWriteProtectChoice(self):
        return bool(self.wpVar.get())

    # return the logical current working directory in a printable form
    # ie. without all the X/.. pairs. The easiest way to do this is to
    # chdir to cwd and get the path there.

    def cwd_print(self):
        os.chdir(self.cwd)
        p = os.getcwd()
        os.chdir(self.orig_dir)
        return p
####
#       Class LoadFileDialog
#
#       Purpose
#       -------
#
#       Specialisation of FileDialog for loading files.
####

class LoadFileDialog(FileDialog):

    def __init__(self, master, title, filter):
        FileDialog.__init__(self, master, title, filter)
        self.top.title(title)

    def OkPressed(self):
        fileName = self.GetFileName()
        if os.path.exists(fileName) == 0:
            msg = 'File ' + fileName + ' not found.'
            errorDlg = alert.ErrorDialog(self.top, msg)
            errorDlg.Show()
            errorDlg.DialogCleanup()
            return
        FileDialog.OkPressed(self)

####
#       Class SaveFileDialog
#
#       Purpose
#       -------
#
#       Specialisation of FileDialog for saving files.
####

class SaveFileDialog(FileDialog):

    def __init__(self, master, title, filter):
        FileDialog.__init__(self, master, title, filter)
        self.top.title(title)

    def OkPressed(self):
        fileName = self.GetFileName()
        if os.path.exists(fileName) == 1:
            msg = 'File ' + fileName + ' exists.\nDo you wish to overwrite it?'
            warningDlg = alert.WarningDialog(self.top, msg)
            if warningDlg.Show() == 0:
                warningDlg.DialogCleanup()
                return
            warningDlg.DialogCleanup()
        FileDialog.OkPressed(self)

#----------------------------------------------------------------------------

#############################################################################
#
# Class:   PersistFileDialog
# Purpose: Essentially the same as FileDialog, except this class contains
#          a class variable (lastAccessedDir) which keeps track of the last
#          directory from which a file was chosen.  Subsequent invocations of
#          this dialog in the same Python session will start up in the last
#          directory where a file was successfully chosen, rather than in the
#          current working directory.
#
# History: M.D. De La Pena, 08 June 2000
#
#############################################################################

class PersistFileDialog(FileDialog):

    # Define a class variable to track the last accessed directory
    lastAccessedDir = None

    def __init__(self, widget, title, filter="*", initWProtState=None):

        FileDialog.__init__(self, widget, title, filter, initWProtState)

        # If the last accessed directory were not None, start up
        # the file browser in the last accessed directory.
        if self.__class__.lastAccessedDir:
            self.cwd      = self.__class__.lastAccessedDir

    # Override the OkPressed method from the parent in order to
    # update the class variable.
    def OkPressed(self):
        self.__class__.lastAccessedDir = self.cwd_print()
        self.TerminateDialog(1)


#############################################################################
#
# Class:   PersistLoadFileDialog
# Purpose: Essentially the same as LoadFileDialog, except this class invokes
#          PersistFileDialog instead of FileDialog.
#
# History: M.D. De La Pena, 08 June 2000
#
#############################################################################

class PersistLoadFileDialog(PersistFileDialog):

    def __init__(self, master, title, filter):
        PersistFileDialog.__init__(self, master, title, filter)
        self.top.title(title)

    def OkPressed(self):
        fileName = self.GetFileName()
        if os.path.exists(fileName) == 0:
            msg = 'File ' + fileName + ' not found.'
            errorDlg = alert.ErrorDialog(self.top, msg)
            errorDlg.Show()
            errorDlg.DialogCleanup()
            return
        PersistFileDialog.OkPressed(self)


#############################################################################
#
# Class:   PersistSaveFileDialog
# Purpose: Essentially the same as SaveFileDialog, except this class invokes
#          PersistFileDialog instead of FileDialog.
#
#############################################################################

class PersistSaveFileDialog(PersistFileDialog):

    def __init__(self, master, title, filter, initWProtState=None):
        PersistFileDialog.__init__(self, master, title, filter, initWProtState)
        self.top.title(title)

    def OkPressed(self):
        fileName = self.GetFileName()
        if os.path.exists(fileName) == 1:
            msg = 'File ' + fileName + ' exists.\nDo you wish to overwrite it?'
            warningDlg = alert.WarningDialog(self.top, msg)
            if warningDlg.Show() == 0:
                warningDlg.DialogCleanup()
                return
            warningDlg.DialogCleanup()
        PersistFileDialog.OkPressed(self)
